#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 24 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

# from ..core.common import ViewerUtils
# from ..core.viewer_work_environment import ViewerWorkEnv
from ..core import ViewerUtils, ViewerWorkEnv
from ...pyqtgraphCore.imageview import ImageView
from ..main_window import MainWindow as ViewerWindow
from ...common import get_sys_config, get_timestamp_str, get_window_manager
from ...common.qdialogs import *
from ...common.utils import make_runfile, make_workdir
from .pytemplates.batch_manager_pytemplate import *
import json
import pandas
from .batch_run_modules import * # DO NOT REMOVE THIS LINE
import uuid
import numpy as np
# from .common import BatchRunInterface
import pickle
import tifffile
import os
from stat import S_IEXEC
# from multiprocessing import Queue
from functools import partial
from collections import deque
import psutil
import traceback
from ...misc_widgets.list_widget_dialog import ListWidgetDialog
from glob import glob
from collections import UserList
from typing import *
from pprint import pformat
import re
from ...common.configuration import IS_WINDOWS

if not IS_WINDOWS:
    from signal import SIGKILL

elif IS_WINDOWS:
    from win32api import TerminateProcess, CloseHandle

import logging


logger = logging.getLogger()


class ModuleGUI(QtWidgets.QWidget):
    """GUI for the Batch Manager"""
    listwchanged = QtCore.pyqtSignal()

    def __init__(self, parent, run_batch: list = None, testing: bool = False):
        logger.info('starting batch mananger')
        QtWidgets.QWidget.__init__(self, parent)
        self._testing = testing
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        if not IS_WINDOWS:
            self.ui.checkBoxUseWorkDir.setChecked(True)
            self.ui.checkBoxUseWorkDir.toggled.connect(self.set_workdir)
        else:
            self.ui.checkBoxUseWorkDir.setChecked(False)
            self.ui.checkBoxUseWorkDir.setDisabled(True)

        self._use_workdir = False  # by default false, set_workdir() performs a check later
        self.working_dir = None
        self.batch_path = None

        self.ui.listwBatch.itemDoubleClicked.connect(self.on_list_widget_batch_doubleclicked)

        self.ui.listwBatch.currentRowChanged.connect(self.ui.listWidgetItemNumbers.setCurrentRow)
        self.ui.listWidgetItemNumbers.currentRowChanged.connect(self.ui.listwBatch.setCurrentRow)

        self.ui.listwBatch.verticalScrollBar().valueChanged.connect(self.ui.listWidgetItemNumbers.verticalScrollBar().setValue)
        self.ui.listWidgetItemNumbers.verticalScrollBar().valueChanged.connect(self.ui.listwBatch.verticalScrollBar().setValue)

        self.ui.listWidgetItemNumbers.verticalScrollBar().setVisible(False)

        self.ui.btnStart.clicked.connect(self.process_batch)
        self.ui.btnStart.setDisabled(True)
        self.ui.btnStartAtSelection.clicked.connect(lambda: self.process_batch(
            start_ix=self.ui.listwBatch.indexFromItem(self.ui.listwBatch.currentItem()).row()))
        self.ui.btnStartAtSelection.setDisabled(True)

        self.ui.btnAbort.clicked.connect(self._terminate_qprocess)
        self.ui.btnAbort.setDisabled(True)

        self.ui.btnAbort_batch.clicked.connect(self.abort_batch)
        self.ui.btnAbort_batch.setDisabled(True)

        self.ui.btnOpen.clicked.connect(self.open_batch)
        self.ui.btnDelete.clicked.connect(self.del_item)
        self.ui.btnViewInput.clicked.connect(self.btn_view_input_slot)
        self.ui.btnNew.clicked.connect(self.ask_create_new_batch)

        listwmodel = self.ui.listwBatch.model()
        listwmodel.rowsInserted.connect(self.listwchanged.emit)
        listwmodel.rowsRemoved.connect(self.listwchanged.emit)

        self.ui.scrollAreaStdOut.setStyleSheet('background-color: #131926')
        self.ui.textBrowserStdOut.setTextColor(QtGui.QColor('#b7b7b7'))

        self.ui.scrollAreaStdOut.hide()
        self.resize(1200, 650)

        self.ui.listwBatch.currentItemChanged.connect(self.show_item_info)
        self.ui.listwBatch.itemClicked.connect(self.show_item_info)

        self.output_widgets = []
        self.df = pandas.DataFrame() #: pandas.DataFrame that stores a "database" of information on the batch

        self.move_processes = []

        self.ui.btnExportShScripts.clicked.connect(self.export_submission_scripts)

        self.lwd = None

        self.ui.lineEditFindItem.textEdited.connect(self.higlight_items)
        self.previous_list_widget_colors = None
        
        self.init_batch(run_batch)

    def get_selected_items(self) -> Tuple[List[int], List[uuid.UUID]]:
        """
        Returns a list of numeric indices and uuids for the currently selected items
        """
        items = self.ui.listwBatch.selectedIndexes()
        indices = [i.row() for i in items]
        uuids = [i.data(3) for i in items]

        return indices, uuids

    def higlight_items(self):
        self.reset_list_widget_colors()
        txt = self.ui.lineEditFindItem.text()
        if not txt:
            return

        items = self.ui.listwBatch.findItems(txt, QtCore.Qt.MatchContains)

        for item in items:
            item.setBackground(QtGui.QBrush(QtGui.QColor('#FF94F7')))

    def init_batch(self, run_batch):
        if not self._testing:
            if run_batch is None:
                path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Choose the location of an existing batch folder '
                                                                        'or a location for a new batch folder')
                if path == '':
                    # does a check to see if the workdir is writable upon batch manager initiation
                    self.set_workdir(True)
                    return
            else:
                path = run_batch[0]
                logger.info('Opening batch: ' + path)

            dfpath = os.path.join(path, 'dataframe.batch')
            if os.path.isfile(dfpath):
                self.open_batch_dir(path)

            else:
                self.create_new_batch_dialog(path)

        self.ui.btnStart.setEnabled(True)
        self.ui.btnStartAtSelection.setEnabled(True)
        self.set_workdir(True)  # does a check to see if the workdir is writable upon batch manager initiation

        if run_batch is not None:
            logger.info('Running from item ' + run_batch[1])
            # ix = self.df.index[self.df['uuid'] == uuid.UUID(run_batch[1])]
            # i = int(ix.to_native_types()[0])
            i = self.get_item_index(run_batch[1])
            self.process_batch(start_ix=i, clear_viewers=True)

    def get_item_index(self, u: Union[uuid.UUID, str]) -> int:
        """
        Get DataFrame index from UUID

        :param u: UUID or str representing UUID
        :return:  numerical index of the DataFrame corresponding to the UUID
        """
        ix = self.df.index[self.df['uuid'] == uuid.UUID(u)]
        i = int(ix.to_native_types()[0])
        return i

    def ask_create_new_batch(self):
        if self.ui.listwBatch.count() > 0:
            if QtWidgets.QMessageBox.warning(self, 'Create Batch?', 'Close the current batch and open another one?',
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
                return
        path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Choose location for a batch')
        if path == '':
            return

        if any(s in path for s in [' ', '(', ')', '?']):
            QtWidgets.QMessageBox.warning(self, 'Invalid path',
                                          'Batch path cannot contain spaces or special characters')
            return

        self.create_new_batch_dialog(path)

    def create_new_batch_dialog(self, path: str):
        name, start = QtWidgets.QInputDialog.getText(self, '', 'Batch Name:', QtWidgets.QLineEdit.Normal, '')

        if not re.match("^[A-Za-z0-9_-]*$", name):
            QtWidgets.QMessageBox.warning(self, 'Invalid name',
                                          'Batch name can only contain alphanumeric characters')
            return

        if start and name != '':
            batch_path = os.path.join(path, name)
            self.create_new_batch(batch_path)

    def create_new_batch(self, full_path: str):
        self.batch_path = full_path
        os.makedirs(self.batch_path)
        self.ui.listwBatch.clear()

        self.df = pandas.DataFrame(
            columns=[
                'module',
                'input_params',
                'output',
                'info',
                'uuid',
                'save_temp_files'
            ]
        )
        self.df.to_pickle(os.path.join(self.batch_path, 'dataframe.batch'))

        self.setWindowTitle('Batch Manager: ' + os.path.basename(self.batch_path))
        self.ui.labelBatchPath.setText(self.batch_path)
        self.show()

    def btn_view_input_slot(self):
        s = self.ui.listwBatch.currentItem()
        UUID = s.data(3)

        r = self.df.loc[self.df['uuid'] == UUID]

        viewers = get_window_manager().viewers

        if len(viewers) > 1:
            self.lwd = ListWidgetDialog()
            self.lwd.listWidget.addItems([str(i + 1) for i in range(len(viewers))])
            self.lwd.label.setText('Viewer to show input in:')
            self.lwd.btnOK.clicked.connect(partial(self.load_item_input, viewers, r, UUID))
        else:
            self.load_item_input(viewers[0], r)

    @present_exceptions('Cannot load input', 'The following occurred when trying to load the input')
    def load_item_input(self, viewers: Union[ViewerWindow, UserList], r: pandas.Series = None, UUID: uuid.UUID = None, *args):
        """
        Pass either the batch DataFrame row or UUID of the item of which to load the input into a viewer

        :param viewers: ViewerWindow or list of ImageView
        :param  r:    Row of batch DataFrame corresponding to the selected item
        :param UUID:  UUID of the item to load input from
        """

        if (r is None) and (UUID is None):
            raise TypeError('Must pass either one of batch item UUID or DataFrame row as Series object')
        elif r is None:
            r = self.df.loc[self.df['uuid'] == UUID]
        elif UUID is None:
            UUID = r['uuid']
            if isinstance(UUID, pandas.Series):
                UUID = UUID.item()

        if isinstance(viewers, ViewerWindow):
            viewer = viewers.viewer_reference
        elif isinstance(viewers, UserList):
            if self.lwd.listWidget.currentItem() is None:
                QtWidgets.QMessageBox.warning(self, 'Nothing selected', 'You must select from the list')
                return
            i = int(self.lwd.listWidget.currentRow())
            viewer = viewers[i].viewer_reference
        else:
            raise TypeError('Must pass pyqtgraphCore.ImageView instance or list of pyqtgraphCore.ImageView instances.')

        vi = ViewerUtils(viewer_reference=viewer)

        if not vi.discard_workEnv():
            return

        # try:
        pikpath = os.path.join(self.batch_path, str(UUID) + '_input.pik')
        tiffpath = os.path.join(self.batch_path, str(UUID) + '_input.tiff')
        vi.viewer.status_bar_label.showMessage('Please wait, loading input into work environment...')
        if os.path.isfile(pikpath) and os.path.isfile(tiffpath):
            vi.viewer.workEnv = ViewerWorkEnv.from_pickle(pickle_file_path=pikpath, tiff_path=tiffpath)
            vi.update_workEnv()
            vi.enable_ui(True)
            vi.viewer.status_bar_label.showMessage('Done! loaded input into work environment.')
            vi.viewer.ui.label_curr_img_seq_name.setText('Input of item: ' + r['name'].item())

        # except:
        #         QtWidgets.QMessageBox.warning(self, 'Input file does not exist',
        #                                       'The input files do not exist for this item.')
        #         vi.viewer.status_bar_label.showMessage('Error, could not load input into work environment.')
        if self.lwd is not None:
            self.lwd.close()
            self.lwd = None

    @present_exceptions('Cannot load output', 'The following error occurred when trying to load the output')
    def on_list_widget_batch_doubleclicked(self, s: QtWidgets.QListWidgetItem):
        self.ui.scrollAreaOutputInfo.show()
        UUID = s.data(3)

        r = self.df.loc[self.df['uuid'] == UUID]

        m = r['module'].item()
        # m = globals()[module]

        output = self.get_batch_item_output(UUID)

        if output is None:
            return

        if output['status'] == 1:
            viewers = get_window_manager().viewers

            if len(viewers) == 0:
                raise IndexError("No Viewers are currently open, you must open a Viewer to view outputs.")

            if len(get_window_manager().viewers) > 1:
                self.lwd = ListWidgetDialog()
                self.lwd.listWidget.addItems([str(i + 1) for i in range(len(viewers))])
                self.lwd.label.setText('Viewer to use for output:')
                self.lwd.btnOK.clicked.connect(partial(self.load_item_output, m, viewers, UUID))
            else:
                self.load_item_output(m, viewers[0], UUID)

    def load_item_output(self, module: str, viewers: Union[ViewerWindow, ImageView, UserList], UUID: uuid.UUID):
        """
        Calls subclass of BatchRunInterface.show_output()

        :param module:  The module name under /batch_run_modules that the batch item is from
        :param viewers: ViewerWindow, ImageView, or list of ViewerWindows
        :param UUID:    UUID of the item to load output from
        """
        if len(self.output_widgets) > 3:
            try:
                w = self.output_widgets.pop(0)
                w.deleteLater()
            except:
                pass

        module = globals()[module]

        if isinstance(viewers, ViewerWindow):
            viewer = viewers.viewer_reference
        elif isinstance(viewers, ImageView):
            viewer = viewers
        elif isinstance(viewers, UserList):
            if self.lwd.listWidget.currentItem() is None:
                QtWidgets.QMessageBox.warning(self, 'Nothing selected', 'You must select from the list')
                return
            i = int(self.lwd.listWidget.currentRow())
            viewer = viewers[i].viewer_reference
        else:
            raise TypeError('"viewers" argument must be one of: ViewerWindow, ImageView, or list of ViewerWindows')

        try:
            self.output_widgets.append(module.Output(self.batch_path, UUID, viewer))
        except:
            QtWidgets.QMessageBox.warning(self, 'Error showing item output',
                                          'The following error occured while '
                                          'trying to load the output of the chosen item\n' + traceback.format_exc())
        if self.lwd is not None:
            self.lwd.close()
            self.lwd = None

    def show_item_info(self, s: QtWidgets.QListWidgetItem):
        """Shows any info (such as the batch module's params) in the meta-info label"""
        if not isinstance(s, QtWidgets.QListWidgetItem):
            return
        UUID = s.data(3)
        row = self.df.loc[self.df['uuid'] == UUID]
        meta = row['info'].item()

        is_pos = lambda x: 1 if x > 0 else 0

        # Format a dict to a nicely readable form
        format_key = lambda d, t: "\n"*is_pos(t) + \
                                  "\n".join(
                                      [": ".join(["\t" * t + k, format_key(v, t + 1)]) for k, v in d.items()]
                                  ) if isinstance(d, dict) else str(d)

        info = format_key(meta, 0)

        self.ui.textBrowserItemInfo.setText(str(UUID) + '\n\n' + info)

        output = self.get_batch_item_output(UUID)

        self.ui.textBrowserOutputInfo.setText('')

        if output is None:
            self.ui.textBrowserOutputInfo.setText('Output file does not exist for selected item')
            return
        else:
            self.ui.textBrowserOutputInfo.setText(format_key(output, 0))

    def disable_ui_buttons(self, b):
        self.ui.btnStart.setDisabled(b)
        self.ui.btnStartAtSelection.setDisabled(b)
        self.ui.btnDelete.setDisabled(b)
        self.ui.btnAbort.setEnabled(b)
        self.ui.btnAbort_batch.setEnabled(b)

    def process_batch(self, start_ix: Union[int, uuid.UUID] = 0, clear_viewers=False):
        """Process everything in the batch by calling subclass of BatchRunInterface.process() for all items in batch

        :param start_ix:       Either DataFrame index (int) or UUID of the item to start from.
        :param clear_viewers:  Clear work environments in all viewers that are open
        """

        if isinstance(start_ix, (str, uuid.UUID)):
            try:
                ix = uuid.UUID(start_ix)
                start_ix = self.get_item_index(ix)
            except:
                raise TypeError('Argument stat_ix only accepts types int, UUID, or str representing a UUID')

        self.ui.checkBoxUseWorkDir.setDisabled(True)

        if len(self.df.index) == 0:
            return

        if not clear_viewers:
            if QtWidgets.QMessageBox.question(self, 'Clear all viewers?',
                                                  'Would you like to clear all viewer work '
                                                  'environments before starting the batch?',
                                                  QtWidgets.QMessageBox.No,
                                                  QtWidgets.QMessageBox.Yes) == QtWidgets.QMessageBox.Yes:
                self._clear_viewers()
        else:
            self._clear_viewers()

        self.current_batch_item_index = start_ix - 1
        self.disable_ui_buttons(True)
        # self.run_next_item()
        self.ui.scrollAreaStdOut.show()
        self.ui.scrollAreaOutputInfo.show()
        self.current_std_out = deque(maxlen=100)
        self.run_batch_item()

    def set_list_widget_item_color(self, ix: int, color: str):
        if color == 'orange':
            self.ui.listwBatch.item(ix).setBackground(QtGui.QBrush(QtGui.QColor('#ffb347')))
        elif color == 'green':
            self.ui.listwBatch.item(ix).setBackground(QtGui.QBrush(QtGui.QColor('#77dd77')))
        elif color == 'red':
            self.ui.listwBatch.item(ix).setBackground(QtGui.QBrush(QtGui.QColor('#fe0d00')))
        elif color == 'blue':
            self.ui.listwBatch.item(ix).setBackground(QtGui.QBrush(QtGui.QColor('#85e3ff')))

    def _clear_viewers(self):
        for viewer in get_window_manager().viewers:
            vi = ViewerUtils(viewer.viewer_reference)
            vi.discard_workEnv()

    @QtCore.pyqtSlot()
    def run_batch_item(self):
        if self.current_batch_item_index > -1:
            UUID = self.df.iloc[self.current_batch_item_index]['uuid']
            output = self.get_batch_item_output(UUID)

            # Deal with the previous batch item if it aborted and perform workdir cleanup
            if output is None:
                self.set_list_widget_item_color(ix=self.current_batch_item_index, color='orange')
                if self._use_workdir:
                    # cleanup workdir
                    self.move_files([], UUID)

            # Deal with the previous batch item that just finished
            elif output['status']:
                if 'output_files' in output.keys() and self._use_workdir:# and os.path.isdir(self.working_dir):
                    output_files_list = output['output_files']

                    mp = self.move_files(output_files_list, UUID)
                    self.set_list_widget_item_color(ix=self.current_batch_item_index, color='blue')
                    mp.finished.connect(partial(self.set_list_widget_item_color, self.current_batch_item_index, 'green'))

                else:
                    self.set_list_widget_item_color(ix=self.current_batch_item_index, color='green')
            else:
                self.set_list_widget_item_color(ix=self.current_batch_item_index, color='red')
                if self._use_workdir:
                    # cleanup workdir
                    self.move_files([f'{UUID}.out'], UUID)

        self.current_batch_item_index += 1
        self.ui.progressBar.setValue(int(self.current_batch_item_index / len(self.df.index) * 100))

        if self.current_batch_item_index == len(self.df.index):
            self.batch_finished()
            return

        r = self.df.iloc[self.current_batch_item_index]

        self.process = QtCore.QProcess()
        self.process.setProcessChannelMode(QtCore.QProcess.MergedChannels)
        # self.process.readyReadStandardError.connect(partial(self.print_qprocess_std_err, self.process))
        self.process.readyReadStandardOutput.connect(partial(self.print_qprocess_std_out, self.process))

        self.process.finished.connect(self.run_batch_item)

        sh_file = self.create_runscript(r, cp=True, mv=False, use_subdir=False)

        self.process.setWorkingDirectory(self.working_dir)

        if IS_WINDOWS:
            self.process.start('powershell.exe', [sh_file])
        else:
            self.process.start(sh_file)
        self.ui.listwBatch.item(self.current_batch_item_index).setBackground(QtGui.QBrush(QtGui.QColor('yellow')))

    def move_files(self, files: list, UUID) -> QtCore.QProcess:
        shell_str = '#!/bin/bash\n'

        files += [f'{UUID}.out']

        for f in files:
            src = os.path.join(self.working_dir, f)
            dst = os.path.join(self.batch_path, f)
            shell_str += f'mv {src} {dst}\n'
        u = f'*{UUID}*'
        shell_str += f'rm {os.path.join(self.working_dir, u)}'
        move_file = os.path.join(self.working_dir, 'move.sh')

        with(open(move_file, 'w')) as sh_mv_f:
            sh_mv_f.write(shell_str)

        mv_st = os.stat(move_file)
        os.chmod(move_file, mv_st.st_mode | S_IEXEC)

        move_process = QtCore.QProcess()
        move_process.setWorkingDirectory(self.working_dir)
        move_process.start(move_file)
        self.move_processes.append(move_process)
        move_process.finished.connect(partial(self.move_processes.remove, move_process))

        return move_process

    def batch_finished(self):
        self.ui.progressBar.setValue(100)
        self.disable_ui_buttons(False)
        self.ui.checkBoxUseWorkDir.setEnabled(True)
        QtWidgets.QMessageBox.information(self, 'Batch is done!', 'Yay, your batch has finished processing!')

    def set_workdir(self, ev):
        if IS_WINDOWS:  # I'm sorry ¯\_(ツ)_/¯
            self.working_dir = self.batch_path
            self._use_workdir = False
            return

        if ev:
            try:
                self.working_dir = make_workdir('batch_manager')
            except Exception as e:
                QtWidgets.QMessageBox.warning(self, 'Cannot create Work Dir',
                                              f'Could not create a work directory, '
                                              f'the batch directory will be used for work.\n{e}')

                self.ui.checkBoxUseWorkDir.setChecked(False)
                self.working_dir = self.batch_path
            else:
                self._use_workdir = True

        else:
            self.working_dir = self.batch_path
            self._use_workdir = False

    def create_runscript(self, r, cp: bool, mv: bool, use_subdir: bool = True) -> str:
        m = globals()[r['module']]
        module_path = os.path.abspath(m.__file__)
        u = r['uuid']

        # Just disable seperation of batch dir and work dir on windows for now
        if IS_WINDOWS:
            cp = False
            mv = False

        if cp:
            files = os.path.join(self.batch_path, f"*{u}*")
            cp_str = f'cp {files} {self.working_dir}\nexport CURR_BATCH_DIR={self.batch_path}'
        else:
            cp_str = None

        if mv:
            work_files = os.path.join(self.working_dir, f'*{u}*')
            mv_str = '\n'.join([f'mv -n {work_files} {self.batch_path}', f'rm *{u}*'])
        else:
            mv_str = None

        save_temp_files = r['save_temp_files']

        args = f'"{self.working_dir}" "{u}" {save_temp_files}'
        if use_subdir:
            savedir = os.path.join(self.batch_path, f'jobs_{get_timestamp_str()}')
        else:
            savedir = self.batch_path

        if not IS_WINDOWS:
            extn = '.sh'
        else:
            extn = '.ps1'

        return make_runfile(module_path=module_path,
                            savedir=savedir,
                            args_str=args,
                            filename=f'{u}{extn}',
                            pre_run=cp_str,
                            post_run=mv_str)

    def export_submission_scripts(self):
        to_copy = self.ui.checkBoxUseWorkDir.isChecked()
        to_move = to_copy
        for ix, r in self.df.iterrows():
            self.create_runscript(r, cp=to_copy, mv=to_move)

        QtWidgets.QMessageBox.information(self, 'Exported',
                                          'Submission scripts for this batch '
                                          'have been exported to a jobs dir in your batch dir')

    def get_batch_item_output(self, UUID: uuid.UUID):
        if self.working_dir is not None:
            out_file = os.path.join(self.working_dir, f'{UUID}.out')
            if os.path.isfile(out_file):
                output = json.load(open(out_file, 'r'))
                return output

        out_file = os.path.join(self.batch_path, f'{UUID}.out')

        if os.path.isfile(out_file):
            output = json.load(open(out_file, 'r'))
            return output
        else:
            return None

    def _terminate_qprocess(self):
        try:
            py_proc = psutil.Process(self.process.pid()).children()[0].pid
        except psutil.NoSuchProcess:
            return
        children = psutil.Process(py_proc).children()

        if not IS_WINDOWS:
            os.kill(py_proc, SIGKILL)

            for child in children:
                os.kill(child.pid, SIGKILL)
        
        if IS_WINDOWS:
            TerminateProcess(py_proc, -1)
            CloseHandle(py_proc)

            for child in children:
                TerminateProcess(child.pid, -1)
                CloseHandle(child.pid)

    def abort_batch(self):
        if QtWidgets.QMessageBox.warning(
                self,
                'Abort batch?',
                'Are you sure you want to abort the batch run?',
                QtWidgets.QMessageBox.Yes,
                QtWidgets.QMessageBox.No
        ) == QtWidgets.QMessageBox.No:
            return

        self.process.finished.disconnect()  # stops it from going to the next item

        self.set_list_widget_item_color(ix=self.current_batch_item_index, color='orange')

        self._terminate_qprocess()  # terminate the qproess for current batch item
        self.disable_ui_buttons(False)
        self.ui.checkBoxUseWorkDir.setEnabled(True)

    def print_qprocess_std_out(self, proc):
        text = proc.readAllStandardOutput().data().decode('utf8')
        # self.current_std_out.append(text)
        self.ui.textBrowserStdOut.append(text)

    def add_item(self, module: str, input_workEnv: ViewerWorkEnv,
                 input_params: dict, name: str = '', info: dict = '') -> uuid.UUID:
        """
        Add an item to the currently open batch

        :param  module:         The module to run from /batch_run_modules.
        :param  input_workEnv:  Input workEnv that the module will use
        :param  input_params:   Input params that the module will use. Depends on your subclass of BatchRunInterface.process() method
        :param  name:           A name for the batch item
        :param  info:           A dictionary with any metadata information to display in the scroll area label.
        :return:                UUID of the added item
        """

        if input_workEnv.isEmpty:
            QtWidgets.QMessageBox.warning(self, 'Work Environment is empty!', 'The current work environment is empty,'
                                                                              ' nothing to add to the batch!')
            return

        UUID = uuid.uuid4()

        filename = os.path.join(self.batch_path, f'{UUID}_input')
        input_workEnv.to_pickle(
            self.batch_path,
            filename=filename,
            save_img_seq=True,
            UUID=UUID
        )

        pickle.dump(input_params, open(os.path.join(self.batch_path, str(UUID) + '.params'), 'wb'), protocol=4)

        input_params = np.array(input_params, dtype=object)

        self.df = self.df.append(
            {
                'module':           module,
                'name':             name,
                'input_item':       None,
                'input_params':     input_params,
                'info':             info,
                'uuid':             UUID,
                'output':           None,
                'save_temp_files':  0
            },
            ignore_index=True
        )

        self.ui.listwBatch.addItem(module + ': ' + name)
        n = self.ui.listwBatch.count()
        item = self.ui.listwBatch.item(n - 1)
        assert isinstance(item, QtWidgets.QListWidgetItem)
        item.setData(3, UUID)
        self.set_line_numbers()

        self.df.to_pickle(os.path.join(self.batch_path, 'dataframe.batch'))
        return UUID

    def del_item(self):
        """Delete the currently selected item from the batch and any corresponding dependents of the item's output"""
        if QtWidgets.QMessageBox.question(self, 'Confirm deletion',
                                          'Are you sure you want to delete the selected items from the batch? '
                                          'This will also remove ALL files associated to the item',
                                          QtWidgets.QMessageBox.Yes,
                                          QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
            return
        items = self.ui.listwBatch.selectedItems()

        for s in items:
            UUID = s.data(3)

            self.df = self.df[self.df['uuid'] != UUID]
            self.df.reset_index(drop=True, inplace=True)

            ix = self.ui.listwBatch.indexFromItem(s).row()
            self.ui.listwBatch.takeItem(ix)
            self.set_line_numbers()

            for file in glob(self.batch_path + '/*' + str(UUID) + '*'):
                os.remove(file)

        self.df.to_pickle(self.batch_path + '/dataframe.batch')

    def set_line_numbers(self):
        self.ui.listWidgetItemNumbers.clear()
        items = list(map(str, range(self.df.index.size)))
        self.ui.listWidgetItemNumbers.addItems(items)

        self.ui.listWidgetItemNumbers.setCurrentRow(self.ui.listwBatch.currentRow())

    def save_batch(self):
        path = QtWidgets.QFileDialog.getSaveFileName(self, 'Save Batch as', '', '(*.batch)')
        if path == '':
            return

        if path[0].endswith('.batch'):
            path = path[0]
        else:
            path = path[0] + '.batch'

        try:
            d = {'batch_path': self.batch_path,
                 'df': self.df}
            pickle.dump(d, open(path[0]))
        except Exception as e:
            QtWidgets.QMessageBox.warning(self, 'File save Error', 'Unable to save the file\n' + str(e))

    def open_batch(self):
        if self.ui.listwBatch.count() > 0:
            if QtWidgets.QMessageBox.warning(self, 'Open Batch', 'Close the current batch and open a new one?',
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
                return
        path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Open Batch', directory='/home/')
        if path == '':
            return

        if any(s in path for s in [' ', '(', ')', '?']):
            QtWidgets.QMessageBox.warning(self, 'Invalid batch path', 'Batch path can only contain alphanumeric characters')
            return
        self.open_batch_dir(path)

    def open_batch_dir(self, path: str):
        dfpath = path + '/dataframe.batch'
        if not os.path.isfile(dfpath):
            QtWidgets.QMessageBox.warning(self, 'Invalid batch dir',
                                          'The selected directory does not appear to be a valid  batch directory '
                                          'since it does not contain a "dataframe.batch" file')
            return
        try:
            df = pandas.read_pickle(dfpath)
            self.df = df
            self.batch_path = path
            self.setWindowTitle('Batch Manager: ' + os.path.basename(self.batch_path))
            self.ui.labelBatchPath.setText(self.batch_path)

            self.ui.listwBatch.clear()

            for ix, r in self.df.iterrows():
                self.ui.listwBatch.addItem(r['module'] + ': ' + r['name'])
                n = self.ui.listwBatch.count()
                item = self.ui.listwBatch.item(n - 1)
                item.setData(3, r['uuid'])

                output = self.get_batch_item_output(r['uuid'])

                if output is None:
                    continue
                elif output['status']:
                    self.ui.listwBatch.item(n - 1).setBackground(
                        QtGui.QBrush(QtGui.QColor('#77dd77'))) # green
                else:
                    self.ui.listwBatch.item(n - 1).setBackground(
                        QtGui.QBrush(QtGui.QColor('#fe0d00'))) # red

            self.set_line_numbers()

        except Exception as e:
            QtWidgets.QMessageBox.warning(self, 'File open Error!',
                                          'Could not open the dataframe file.\n' + traceback.format_exc())
            return

        self.disable_ui_buttons(False)

    def reset_list_widget_colors(self):
        for ix, r in self.df.iterrows():
            item = self.ui.listwBatch.item(ix)

            output = self.get_batch_item_output(r['uuid'])

            if output is None:
                item.setBackground(QtGui.QBrush(QtGui.QColor('#FFFFFF')))
            elif output['status']:
                item.setBackground(QtGui.QBrush(QtGui.QColor('#77dd77')))  # green
            else:
                item.setBackground(QtGui.QBrush(QtGui.QColor('#fe0d00')))  # red



from bokeh.models import ColumnDataSource
from typing import *
import pandas
from uuid import UUID
import inspect
import logging


logger = logging.getLogger()


class BokehCallbackSignal:
    def __init__(
        self,
        dataframe: Optional[pandas.DataFrame] = None,
        source: Optional[ColumnDataSource] = None
    ):
        """

        :param dataframe: associate a DataFrame with this signal
                          ``dataframe`` is generally a superset of ``source``

        :param source:  associate a ColumnDataSource with this signal.
                        the data contained in ``source`` must be a subset of ``dataframe``.
                        ``source`` typically does not contain some columns from ``dataframe``
                        due to incompatibility of the ``ColumnDataSource`` type.
        """
        self.dataframe: Union[pandas.DataFrame] = dataframe
        self.source_data: Union[ColumnDataSource, None] = source

        self.callbacks = []
        self.callbacks_data = []

        self._pause_state = False

    def connect(self, func: callable):
        """
        Use this to just pass the new value from the widget or glyph to the callback function

        :param func:
        :return:
        """
        self.callbacks.append(func)

    def connect_data(self, func: callable, identifier: str):
        """
        Use this to pass a sub dataframe to the callback function.

        ``self.dataframe`` and

        :param func:
        :return:
        """

        self.callbacks_data.append((func, identifier))

    def disconnect(self, func: callable):
        if func in self.callbacks:
            self.callbacks.remove(func)
        elif func in self.callbacks_data:
            self.callbacks_data.remove(func)

    def pause(self):
        """
        Disable callback trigger
        """
        self._pause_state = True

    def unpause(self):
        """
        Enable callback trigger
        """
        self._pause_state = False

    def trigger(self, attr, old, val):
        """
        This function signature must have 3 and only 3 args, bokeh is very picky

        :param attr: name of the attribute

        :param old: old value of the attribute

        :param val: the index within `self.source_data` (such as from a glyph using a `ColumnDataSource`)
                    or the new value (such as from a widget)
        :type val:  Union[List[int], Any]

        :return:
        """
        if self._pause_state:
            return

        # if there are multiple values
        # such as if multiple datapoints were selected in a glyph
        if isinstance(val, list):
            if len(val) != 1:
                return

        if self.source_data is not None:
            self._trigger_callbacks_data(val)

        self._trigger_callbacks(val)

    def _trigger_callbacks(self, val: Any):
        """
        Trigger callbacks which directly take a value, such as from a widget

        :param val:
        :return:
        """
        for func in self.callbacks:
            logger.debug(f"VAL IN SIGNAL: {val}")
            if bool(inspect.signature(func).parameters):  # make sure the function takes arguments
                func(val)  # send out val directly
            else:
                func()  # else just call the function

    def _trigger_callbacks_data(self, val: List[int]):
        """
        Trigger the callbacks which take a subdataframe as the argument

        :param val: list containing a single integer, which is an index for source_data
        :return:
        """
        logger.debug(f"VAL IN SIGNAL: {val}")
        for func, identifier in self.callbacks_data:
            uid = self.source_data.data[identifier][val][0]  # get the source_data present at the `val` index
            logger.debug(uid)
            # subdataframe where the identifier UUID matches
            out = self.dataframe[self.dataframe[identifier] == uid]
            logger.debug(out.copy(deep=True))
            func(out)


class WebPlot:
    def __init__(self, *args, **kwargs):
        attrs = dir(self)
        self.signals: List[BokehCallbackSignal] = \
            [getattr(self, a) for a in attrs if isinstance(getattr(self, a), BokehCallbackSignal)]
        logger.debug("WEBPLOT INIT")
        logger.debug(self.signals)

    @classmethod
    def signal_blocker(cls, func):
        """Block callbacks, used when the plot x and y limits change due to user interaction"""

        logger.debug("***** SIGNAL BLOCKER CALLED ******")

        def fn(self, *args, **kwargs):
            logger.debug("self.signals is")
            logger.debug(self.signals)
            for signal in self.signals:
                logger.debug(f"Blocking signal {signal}")
                signal.pause()

            ret = func(self, *args, **kwargs)

            for signal in self.signals:
                signal.unpause()

            return ret
        return fn



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from ..plot_window import PlotWindow
from .beeswarm_plot_controls_pytemplate import Ui_BeeswarmControls
from ...variants import BeeswarmPlot
from .. import DatapointTracerWidget
from ...variants import ViolinsPlot
import numpy as np
from uuid import UUID
import pandas as pd
from ....common.qdialogs import use_save_file_dialog, present_exceptions
import os


class SummaryStatsWidget(QtWidgets.QWidget):
    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self)

        self.vlayout = QtWidgets.QVBoxLayout(self)

        self.text_widget = QtWidgets.QPlainTextEdit(self)
        self.text_widget.setReadOnly(True)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.text_widget.setFont(font)
        self.vlayout.addWidget(self.text_widget)

        self.button_save = QtWidgets.QPushButton(self)
        self.button_save.setText("Save to file")
        self.button_save.clicked.connect(lambda: self.save_to_file())
        self.vlayout.addWidget(self.button_save)

        self.setLayout(self.vlayout)

        self.data: str = ''

    def clear(self):
        self.data = ''
        self.text_widget.clear()

    def set_data(self, s: str):
        self.clear()
        self.data = s
        self.text_widget.setPlainText(self.data)

    @use_save_file_dialog("Save summary stats file", ext='.stats')
    @present_exceptions()
    def save_to_file(self, path, *args):
        if not self.data:
            raise ValueError(
                "Summary stats are empty, nothing to save."
            )
        if os.path.isfile(path):
            raise FileExistsError(
                "A file with that name already exists, specify a different path"
            )

        with open(path, 'w') as f:
            f.write(self.data)


class ControlWidget(QtWidgets.QWidget, Ui_BeeswarmControls):
    signal_changed = QtCore.pyqtSignal(str, dict)

    def __init__(self, parent=None):
        super(ControlWidget, self).__init__(parent)  # parent)
        self.setupUi(parent)

    def apply_all_settings(self):
        data_columns = self.listWidgetDataColumns.selectedItems()
        if self.radioButtonGroupBySingleColumn.isChecked():
            if self.comboBoxGrouping.currentText() == '':
                QtWidgets.QMessageBox.warning(None, 'Invalid parameters', 'Invalid grouping option')
                return
            grouping = {'type': 'column', 'column_name': self.comboBoxGrouping.currentText()}

        elif self.radioButtonGroupByTransmissions.isChecked():
            grouping = {'type': 'transmissions'}

        else:
            QtWidgets.QMessageBox.warning(None, 'Invalid parameters', 'No grouping selected')
            return

        self.signal_changed.emit(data_columns, grouping)


class BeeswarmPlotWindow(PlotWindow):
    def __init__(self, parent=None):
        super(BeeswarmPlotWindow, self).__init__(parent)
        self.setWindowTitle('Beeswarm Plot Window')
        # self.plot_obj = BeeswarmPlot(self.ui.graphicsView, self)
        # self.plot_obj.signal_spot_clicked.connect(print)

        # Setup the control widget. Pass the main control widget to the Beeswarm specific control widget class as its parent so that beeswarm specific stuff is just added to it.
        self.ui.groupBoxSpecific.setLayout(QtWidgets.QVBoxLayout())
        self.control_widget = ControlWidget(self.ui.groupBoxSpecific)
        self.ui.groupBoxSpecific.layout().addWidget(self.control_widget)

        self.add_plot_tab('Beeswarm')
        self.beeswarm_plot = BeeswarmPlot(self.graphicsViews['Beeswarm'])
        # self.add_plot_tab('Violins')
        self.violins_plot = ViolinsPlot()
        self.ui.tabWidget.addTab(self.violins_plot, 'Violins')
        # self.ui.tabWidget.widget(1).setLayout(QtWidgets.QVBoxLayout())

        self.current_datapoint = None

        self.beeswarm_plot.signal_spot_clicked.connect(self.set_current_datapoint)

        # self.control_widget.btnTraceDatapoint.clicked.connect(self.open_datapoint_tracer)
        self.datapoint_tracers = []

        self.live_datapoint_tracer = DatapointTracerWidget()
        self.ui.actionLive_datapoint_tracer.triggered.connect(self.live_datapoint_tracer.show)

        self.summary_stats_widget = SummaryStatsWidget(parent=None)
        self.control_widget.pushButtonSummaryStats.clicked.connect(
            self.summary_stats_widget.show
        )

    def get_current_datapoint(self) -> UUID:
        return self.current_datapoint

    @QtCore.pyqtSlot(UUID)
    def set_current_datapoint(self, identifier: UUID):
        self.current_datapoint = identifier

        tstart = None
        tend = None

        if '_pf_b_ix_l' in self.dataframe.columns:
            tstart = self.dataframe[self.dataframe[self.uuid_column] == str(identifier)]['_pf_b_ix_l']
            if isinstance(tstart, pd.Series):
                tstart = tstart.item()
        if '_pf_b_ix_r' in self.dataframe.columns:
            tend = self.dataframe[self.dataframe[self.uuid_column] == str(identifier)]['_pf_b_ix_r']
            if isinstance(tend, pd.Series):
                tend = tend.item()

        r = self.dataframe[self.dataframe[self.uuid_column] == str(identifier)]

        if isinstance(r._BLOCK_, pd.Series):
            block_id = r._BLOCK_.item()
        elif isinstance(r._BLOCK_, str):
            block_id = r._BLOCK_

        h = self.merged_transmission.history_trace.get_data_block_history(block_id)

        self.live_datapoint_tracer.set_widget(datapoint_uuid=identifier,
                                              data_column_curve=self.datapoint_tracer_curve_column,
                                              row=r,
                                              proj_path=self.merged_transmission.get_proj_path(),
                                              history_trace=h,
                                              tstart=tstart, tend=tend
                                              )

    def update_params(self):
        super(BeeswarmPlotWindow, self).update_params()
        spot_size = self.control_widget.horizontalSliderSpotSize.value()
        self.update_beeswarm()
        self.update_violins()

        summary_stats = self.dataframe.groupby(self.grouping_column).describe().unstack(1).to_string()
        self.summary_stats_widget.set_data(summary_stats)

    def update_beeswarm(self):
        self.beeswarm_plot.clear_plots()
        colors = self.auto_colormap(len(self.groups))

        colors_map = {}
        for ix, g in enumerate(self.groups):
            colors_map.update({g: colors[ix]})

        accepted_datatypes = [int, float, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32,
                              np.uint64, np.float16, np.float32, np.float64, np.uint64, np.float16, np.float32,
                              np.float64]

        if not self.data_types_check(accepted_datatypes):
            return

        for i, data_column in enumerate(self.data_columns):
            pr = int(((i + 1) / len(self.data_columns)) * 100)
            msg = f'Progress: {pr} % ,Plotting data column: {data_column}'
            self.status_bar.showMessage(msg)
            self.beeswarm_plot.add_plot(data_column)

            for ix, (dataframe, group) in enumerate(zip(self.group_dataframes, self.groups)):
                self.status_bar.showMessage(f'{msg}, plotting group: {group}')
                self.beeswarm_plot.add_data_to_plot(i, data_series=dataframe[data_column],
                                                    uuid_series=dataframe[self.uuid_column],
                                                    name=group, color=colors[ix])
        self.beeswarm_plot.set_legend(colors_map)

    def update_violins(self):
        cols = self.data_columns + [self.grouping_column]
        sub_df = self.dataframe.loc[:, cols]
        self.violins_plot.set(sub_df, x_column=self.grouping_column, data_columns=self.data_columns, x_order=self.groups)


    # def open_datapoint_tracer(self):
    #     identifier = self.get_current_datapoint()
    #     self.datapoint_tracers.append(DatapointTracerWidget())
    #
    #     r = self.dataframe[self.dataframe[self.uuid_column] == identifier]
    #
    #     self.datapoint_tracers[-1].set_widget(datapoint_uuid=identifier,
    #                                           data_column_curve=self.datapoint_tracer_curve_column,
    #                                           parent=self,
    #                                           row=r,
    #                                           proj_path=self.merged_transmission.get_proj_path(),
    #                                           history_trace=self.get_history_trace(identifier),
    #                                           )
    #     self.datapoint_tracers[-1].show()

if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    bpw = BeeswarmPlotWindow()
    bpw.show()
    app.exec_()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FlowchartCtrlTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(270, 499)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 2, 0, 2)
        self.gridLayout.setHorizontalSpacing(6)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.fileNameLabel = QtWidgets.QLabel(Form)
        self.fileNameLabel.setObjectName("fileNameLabel")
        self.verticalLayout.addWidget(self.fileNameLabel)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.loadBtn = QtWidgets.QPushButton(Form)
        self.loadBtn.setObjectName("loadBtn")
        self.horizontalLayout.addWidget(self.loadBtn)
        self.saveBtn = FeedbackButton(Form)
        self.saveBtn.setObjectName("saveBtn")
        self.horizontalLayout.addWidget(self.saveBtn)
        self.saveAsBtn = FeedbackButton(Form)
        self.saveAsBtn.setObjectName("saveAsBtn")
        self.horizontalLayout.addWidget(self.saveAsBtn)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.ctrlList = TreeWidget(Form)
        self.ctrlList.setObjectName("ctrlList")
        self.ctrlList.headerItem().setText(0, "1")
        self.ctrlList.header().setVisible(False)
        self.ctrlList.header().setStretchLastSection(False)
        self.verticalLayout.addWidget(self.ctrlList)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.reloadBtn = FeedbackButton(Form)
        self.reloadBtn.setEnabled(False)
        self.reloadBtn.setCheckable(False)
        self.reloadBtn.setFlat(False)
        self.reloadBtn.setObjectName("reloadBtn")
        self.horizontalLayout_2.addWidget(self.reloadBtn)
        self.showChartBtn = QtWidgets.QPushButton(Form)
        self.showChartBtn.setCheckable(True)
        self.showChartBtn.setObjectName("showChartBtn")
        self.horizontalLayout_2.addWidget(self.showChartBtn)
        self.BtnResetView = QtWidgets.QPushButton(Form)
        self.BtnResetView.setObjectName("BtnResetView")
        self.horizontalLayout_2.addWidget(self.BtnResetView)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.fileNameLabel.setText(_translate("Form", "Unsaved"))
        self.loadBtn.setText(_translate("Form", "Load.."))
        self.saveBtn.setText(_translate("Form", "Save"))
        self.saveAsBtn.setText(_translate("Form", "Save As.."))
        self.reloadBtn.setText(_translate("Form", "Reload Libs"))
        self.showChartBtn.setText(_translate("Form", "Flowchart"))
        self.BtnResetView.setText(_translate("Form", "Reset view"))

from ..widgets.FeedbackButton import FeedbackButton
from ..widgets.TreeWidget import TreeWidget



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 26 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

"""

from .tab_page_pytemplate import *
from .row import Row
import pandas as pd
from functools import partial
from typing import Union


class Page(QtWidgets.QWidget):
    def __init__(self, parent, stim_type: str):
        QtWidgets.QWidget.__init__(self)

        self.ui = Ui_TabPage()
        self.ui.setupUi(self)

        self.stim_type = stim_type
        self.rows = []  #: List of row objects, each row holds one stimulus period.
        self.ui.btnAddRow.clicked.connect(lambda: self.add_row())

    def set_data(self, dataframe: pd.DataFrame):
        """
        Set the stimulus map

        :param dataframe: DataFrame with the appropriate rows (see add_row())
        """
        self.clear()
        for ix, series in dataframe.iterrows():
            self.add_row(series)

    def get_dataframe(self) -> pd.DataFrame:
        """
        Get the stimulus map as a DataFrame
        """
        if len(self.rows) < 1:
            raise IndexError('No stimuli input for this stimulus type: ' + self.stim_type)

        l = []
        for row in self.rows:
            l.append(row.get_dict())
        return pd.DataFrame(l)

    def set_units(self, units: str):
        """
        Set the time units

        :param units: One of 'frames' or 'seconds'
        """

        if units not in ['frames', 'seconds']:
            raise ValueError('Units must be either "frames" or "seconds"')

        ix = self.ui.comboBoxTimeUnits.findText(units)
        self.ui.comboBoxTimeUnits.setCurrentIndex(ix)

    def get_units(self) -> str:
        """
        Get the time units
        """
        return self.ui.comboBoxTimeUnits.currentText()

    def add_row(self, pd_series: pd.Series = None):
        """
        Add a row to the stimulus map

        :param pd_series: pandas series containing the following: stimulus name, start, end, and color
        :return:
        """
        row = Row(pd_series)
        self.rows.append(row)
        row.btn_remove.clicked.connect(partial(self.delete_row, row))
        self.ui.verticalLayout.insertLayout(self.ui.verticalLayout.count() - 1, row.hlayout)

    def delete_row(self, row: Union[Row, int]):
        """
        Delete a row from the stimulus map

        :param row: The Row object to remove or the numerical index of the row
        """

        if isinstance(row, int):
            row = self.rows[row]
        if not isinstance(row, Row):
            raise TypeError("row must be either a Row object or int")

        self.ui.verticalLayout.removeItem(row.hlayout)
        self.rows.remove(row)
        row.delete()
        row.hlayout.deleteLater()
        row.deleteLater()

    def clear(self):
        """Clear the stimulus map"""
        while len(self.rows) > 0:
            self.delete_row(self.rows[0])

    def set_stim_autocompleter(self, stimulus_list: list):
        autocompleter = QtWidgets.QCompleter(stimulus_list)
        for row in self.rows:
            row.name.setCompleter(autocompleter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

"""

from PyQt5 import QtWidgets
from ....pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget
from ....pyqtgraphCore.console import ConsoleWidget
from ....pyqtgraphCore.widgets.ComboBox import ComboBox
from ....analysis import Transmission, organize_dataframe_columns, get_proportions
from math import sqrt
from ..base import BasePlotWidget
from ....common.qdialogs import *
from ....common.configuration import console_history_path
from ...utils import auto_colormap, ColormapListWidget
from seaborn import heatmap
import os
from matplotlib.axes import Axes


class ProportionsWidget(BasePlotWidget, MatplotlibWidget):
    drop_opts = ['xs', 'ys']  #: Drop the 'xs' and 'ys' since they are pd.Series objects and not required for restoring the plot

    def __init__(self):
        super().__init__()
        self.message_label = QtWidgets.QLabel()
        self.message_label.setText('')
        self.message_label.setMaximumHeight(30)
        self.vbox.addWidget(self.message_label)

        self.checkbox_update_live = QtWidgets.QCheckBox(self)
        self.checkbox_update_live.setText('Live update from flowchart')
        self.checkbox_update_live.setChecked(True)
        self.checkbox_update_live.toggled.connect(self.set_update_live)
        self.vbox.addWidget(self.checkbox_update_live)

        xs_label = QtWidgets.QLabel(self)
        xs_label.setText('X column')
        xs_label.setMaximumHeight(30)
        self.vbox.addWidget(xs_label)

        self.xs_combo = ComboBox(self)
        self.xs_combo.currentIndexChanged.connect(lambda: self.update_plot())
        self.xs_combo.setMaximumHeight(30)
        self.vbox.addWidget(self.xs_combo)

        ys_label = QtWidgets.QLabel(self)
        ys_label.setText('Y column')
        ys_label.setMaximumHeight(30)
        self.vbox.addWidget(ys_label)

        self.ys_combo = ComboBox(self)
        self.ys_combo.setMaximumHeight(30)
        self.ys_combo.currentTextChanged.connect(lambda: self.update_plot())
        self.vbox.addWidget(self.ys_combo)

        self.checkbox_percent = QtWidgets.QCheckBox(self)
        self.checkbox_percent.setText('Show percentages')
        self.checkbox_percent.setChecked(True)
        self.checkbox_percent.toggled.connect(lambda: self.update_plot())
        self.checkbox_percent.setMaximumHeight(30)
        self.vbox.addWidget(self.checkbox_percent)

        self.radio_barplot = QtWidgets.QRadioButton(self)
        self.radio_barplot.setText('bar plot')
        self.radio_barplot.setChecked(True)
        self.radio_barplot.toggled.connect(lambda: self.update_plot())
        self.radio_barplot.setMaximumHeight(30)

        self.radio_heatmap = QtWidgets.QRadioButton(self)
        self.radio_heatmap.setText('heatmap')
        self.radio_heatmap.toggled.connect(lambda: self.update_plot())
        self.radio_heatmap.toggled.connect(self.show_cmap_listwidget)

        hlayout_radios_btns = QtWidgets.QHBoxLayout()
        hlayout_radios_btns.addWidget(self.radio_barplot)
        hlayout_radios_btns.addWidget(self.radio_heatmap)
        self.vbox.addLayout(hlayout_radios_btns)

        self.cmap_label = QtWidgets.QLabel(self)
        self.cmap_label.setText('Colormap for heatmap')
        self.cmap_label.setMaximumHeight(30)
        self.cmap_label.hide()
        self.vbox.addWidget(self.cmap_label)

        self.cmap_listwidget = ColormapListWidget(self)
        self.cmap_listwidget.set_cmap('viridis')
        self.cmap_listwidget.currentRowChanged.connect(lambda: self.update_plot())
        self.cmap_listwidget.hide()
        self.cmap_listwidget.setMaximumHeight(100)
        self.vbox.addWidget(self.cmap_listwidget)

        hlayout2 = QtWidgets.QHBoxLayout()

        self.btn_update_plot = QtWidgets.QPushButton(self)
        self.btn_update_plot.setText('Update Plot')
        self.btn_update_plot.setMaximumHeight(30)
        self.btn_update_plot.clicked.connect(lambda: self.update_plot())
        hlayout2.addWidget(self.btn_update_plot)

        self.btn_swap_xy = QtWidgets.QPushButton(self)
        self.btn_swap_xy.setText('Swap X-Y')
        self.btn_swap_xy.clicked.connect(self.swap_x_y)
        self.btn_swap_xy.setMaximumHeight(30)
        hlayout2.addWidget(self.btn_swap_xy)

        self.vbox.addLayout(hlayout2)

        hlayout = QtWidgets.QHBoxLayout()

        btn_save = QtWidgets.QPushButton(self)
        btn_save.setText('Save')
        btn_save.clicked.connect(self.save_plot_dialog)
        btn_save.setMaximumHeight(30)
        hlayout.addWidget(btn_save)

        btn_load = QtWidgets.QPushButton(self)
        btn_load.setText('Load')
        btn_load.clicked.connect(self.open_plot_dialog)
        btn_load.setMaximumHeight(30)
        hlayout.addWidget(btn_load)

        self.vbox.addLayout(hlayout)

        hlayout3 = QtWidgets.QHBoxLayout()

        btn_export = QtWidgets.QPushButton(self)
        btn_export.setText('Export CSV')
        btn_export.clicked.connect(self.export)
        btn_export.setMaximumHeight(30)
        hlayout3.addWidget(btn_export)

        btn_show_console = QtWidgets.QPushButton(self)
        btn_show_console.setText('Show Console')
        btn_show_console.setCheckable(True)
        btn_show_console.setChecked(False)
        btn_show_console.setMaximumHeight(30)
        hlayout3.addWidget(btn_show_console)

        self.vbox.addLayout(hlayout3)

        ns = {'this': self,
              'get_fig': lambda: self.fig,
              'get_ax': lambda: self.ax,
              'get_dataframe': lambda: self.props_df,
              }

        txt = ["Namespaces",
               "self as 'this'",
               "get_fig() - get the current Figure instance",
               "get_ax() - get the current Axes instance",
               "get_dataframe() - get the current dataframe of proportions"
               ]

        txt = "\n".join(txt)

        cmd_history_file = os.path.join(console_history_path, 'proportions_plot.pik')

        self.console = ConsoleWidget(parent=self, namespace=ns, text=txt, historyFile=cmd_history_file)
        self.console.setMaximumHeight(300)
        self.console.setVisible(False)

        self.vbox.addWidget(self.console)
        btn_show_console.toggled.connect(self.console.setVisible)

        spacer = QtWidgets.QSpacerItem(0, 10, QtWidgets.QSizePolicy.Expanding)
        self.vbox.insertSpacerItem(0, spacer)

        self.props_df = None

        self.block_signals_list = [self.xs_combo, self.ys_combo, self.checkbox_percent, self.checkbox_update_live]

        self._ax = None

    @property
    def ax(self) -> Axes:
        """
        The Axes object for this plot

        :return: The Axes object for this plot
        :rtype: AXes
        """

        if self._ax is None:
            TypeError("Axes not set")

        return self._ax

    def show_cmap_listwidget(self, b: bool):
        self.cmap_label.setVisible(b)
        self.cmap_listwidget.setVisible(b)

    @BasePlotWidget.signal_blocker
    def set_update_live(self, b: bool):
        self.checkbox_update_live.setChecked(b)
        self.update_live = b

        if b:
            self.update_plot()

    @BasePlotWidget.signal_blocker
    def swap_x_y(self, *args, **kwargs):
        """Swap the X and Y axes"""
        xs_opt = self.xs_combo.currentText()
        ys_opt = self.ys_combo.currentText()

        ix_x = self.xs_combo.findText(ys_opt)
        self.xs_combo.setCurrentIndex(ix_x)

        ix_y = self.ys_combo.findText(xs_opt)
        self.ys_combo.setCurrentIndex(ix_y)
        # self._block = False

        self.update_plot()

    @BasePlotWidget.signal_blocker
    def set_input(self, transmission: Transmission):
        super(ProportionsWidget, self).set_input(transmission)
        if self.update_live:
            self.update_plot()

    @BasePlotWidget.signal_blocker
    def fill_control_widget(self, data_columns: list, categorical_columns: list, uuid_columns: list):
        # self.xs_combo.clear()
        self.xs_combo.setItems(categorical_columns)

        # self.ys_combo.clear()
        self.ys_combo.setItems(categorical_columns)

    def get_plot_opts(self, drop: bool = False):
        """
        Get the plot options

        :param drop: Drop the 'xs' and 'ys' objects when saving the returned dict for saving to an hdf5 file
        """
        xs_name = self.xs_combo.currentText()
        ys_name = self.ys_combo.currentText()

        opts = dict(xs_name=xs_name, ys_name=ys_name,
                    xs=self.transmission.df[xs_name],
                    ys=self.transmission.df[ys_name],
                    percentages=self.checkbox_percent.isChecked(),
                    use_barplot=self.radio_barplot.isChecked(),
                    use_heatmap=self.radio_heatmap.isChecked(),
                    heatmap_cmap=self.cmap_listwidget.current_cmap
                    )
        if drop:
            for k in self.drop_opts:
                opts.pop(k)
        return opts

    @BasePlotWidget.signal_blocker
    def set_plot_opts(self, opts: dict):
        """Set the plot options"""
        ix_x = self.xs_combo.findText(opts['xs_name'])
        self.xs_combo.setCurrentIndex(ix_x)

        ix_y = self.ys_combo.findText(opts['ys_name'])
        self.ys_combo.setCurrentIndex(ix_y)

        self.checkbox_percent.setChecked(opts['percentages'])

        try:
            self.radio_barplot.setChecked(opts['use_barplot'])
            self.radio_heatmap.setChecked(opts['use_heatmap'])
            self.cmap_listwidget.set_cmap(opts['heatmap_cmap'])
        except:
            pass

    @present_exceptions('Plot error', 'Cannot plot, make sure you have selected appropriate data columns')
    def update_plot(self):
        """
        Update the plot data and draw
        """
        # self.ax.cla()
        self.fig.clear()
        self._ax = self.fig.add_subplot(111)  #: The axis that is used for this plot

        opts = self.get_plot_opts()
        use_barplot = opts.pop('use_barplot')
        use_heatmap = opts.pop('use_heatmap')
        heatmap_cmap = opts.pop('heatmap_cmap')

        if opts['xs_name'] == opts['ys_name']:
            self.message_label.setText('Must choose different X and Y columns')
            return
        self.message_label.clear()

        ys = self.transmission.df[opts['ys_name']]

        show_percents = opts['percentages']

        self.props_df = get_proportions(**opts)  #: DataFrame instance containing proportions data according to the plot parameters

        n_labels = len(ys.unique())

        if n_labels < 11:
            cmap = 'tab10'
        elif 10 < n_labels < 21:
            cmap = 'tab20'
        elif 20 < n_labels < 211:
            cmap = 'nipy_spectral'
        else:
            raise ValueError('Cannot generate colormap for greater than 210 labels.\n'
                             'Why would you have > 210 labels?')

        colors = auto_colormap(n_labels, cmap)

        if show_percents:
            label = 'percentages'
        else:
            label = 'raw counts'

        if use_barplot:
            self.props_df.plot(kind='bar', stacked=True, ax=self.ax, color=colors)
            self.ax.set_xlabel(opts['xs_name'])
            self.ax.set_ylabel(label)
            # self.fig.tight_layout()
            self.ax.legend(bbox_to_anchor=(0.0, 1.02, 1.0, 0.102), loc='upper center', ncol=int(sqrt(len(set(ys)))),
                           mode='expand')

        elif use_heatmap:
            props_df = self.props_df.fillna(value=0.)
            heatmap(props_df.T, ax=self.ax, cmap=heatmap_cmap, annot=True, fmt='.1f', cbar_kws={'label': label})
            self.fig.tight_layout()

        self.draw()
        self.toolbar.update()

    @present_exceptions('Export error', 'The following error occurred.')
    def export(self, *args, **kwargs):
        """Export as a csv file"""
        if self.props_df is None:
            return
        if self.transmission is None:
            return

        try:
            proj_path = self.transmission.get_proj_path()
        except ValueError:
            proj_path = ''

        path = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file as', proj_path, '(*.csv)')
        if path == '':
            return

        path = path[0]
        if not path.endswith('.csv'):
            path = f'{path}.csv'

        self.props_df.to_csv(path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 13 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ..core.common import ViewerUtils
from .pytemplates.roi_manager_pytemplate import *
from .roi_manager_modules import managers
from .roi_manager_modules.roi_types import *
from functools import partial
import traceback
import logging


logger = logging.getLogger()


class ModuleGUI(QtWidgets.QDockWidget):
    """The GUI front-end for the ROI Manager module"""
    def __init__(self, parent, viewer_reference):
        """Instantiate attributes"""
        self.vi = ViewerUtils(viewer_reference)

        QtWidgets.QDockWidget.__init__(self, parent)

        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.manager = managers.ManagerManual(self, self.ui, self.vi) #: The back-end manager instance.
        self.vi.viewer.workEnv.roi_manager = self.manager

        self.vi.viewer.workEnv.rois = self.manager.roi_list
        self.vi.viewer.ui.splitter.setEnabled(True)

        action_add_ellipse_roi = QtWidgets.QWidgetAction(self)
        action_add_ellipse_roi.setText('Add Ellipse ROI')
        action_add_ellipse_roi.triggered.connect(partial(self.add_manual_roi, 'EllipseROI'))

        self.btnAddROIMenu = QtWidgets.QMenu(self)
        self.btnAddROIMenu.addAction(action_add_ellipse_roi)

        self.ui.btnAddROI.clicked.connect(partial(self.add_manual_roi, 'PolyLineROI'))
        self.ui.btnAddROI.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.ui.btnAddROI.customContextMenuRequested.connect(self.btnAddROI_context_menu_requested)

        self.ui.btnSwitchToManualMode.clicked.connect(self.start_manual_mode)

        action_delete_roi = QtWidgets.QWidgetAction(self)
        action_delete_roi.setText('Delete')
        action_delete_roi.triggered.connect(self.slot_delete_roi_menu)

        self.list_widget_context_menu = QtWidgets.QMenu(self)
        self.list_widget_context_menu.addAction(action_delete_roi)

        self.ui.listWidgetROIs.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.ui.listWidgetROIs.customContextMenuRequested.connect(self._list_widget_context_menu_requested)

        self.ui.listWidgetROIs.currentRowChanged.connect(self.request_zlevel)

        self.ui.pushButtonImportFromImageJ.clicked.connect(self.import_from_imagej)

        self.installEventFilter(self)

        self.ui.lineEditROITag.sig_key_right.connect(self.play_forward)
        self.ui.lineEditROITag.sig_key_left.connect(self.play_backward)

        self.ui.lineEditROITag.sig_key_home.connect(self.img_seq_home)
        self.ui.lineEditROITag.sig_key_end.connect(self.img_seq_end)

        self.ui.horizontalSliderSpotSize.valueChanged.connect(self.set_spot_size)

        for w in \
            [
                self.ui.radioButton_curve_data,
                self.ui.radioButton_dfof,
                self.ui.radioButton_spikes
            ]:
            w.clicked.connect(self.set_curveplot_datatype)

    def eventFilter(self, QObject, QEvent):
        """Set some keyboard shortcuts"""
        if QEvent.type() == QEvent.KeyPress:

            if len(self.manager.roi_list) < 1:
                return super(ModuleGUI, self).eventFilter(QObject, QEvent)

            max_ix = len(self.manager.roi_list) - 1

            if QEvent.key() == QtCore.Qt.Key_PageUp:
                ix = self.manager.roi_list.current_index
                self.ui.listWidgetROIs.setCurrentRow(max(0, ix - 1))

            elif QEvent.key() == QtCore.Qt.Key_PageDown:
                ix = self.manager.roi_list.current_index
                self.ui.listWidgetROIs.setCurrentRow(min(max_ix, ix + 1))

            elif QEvent.key() == QtCore.Qt.Key_Right:
                self.play_forward()

            elif QEvent.key() == QtCore.Qt.Key_Left:
                self.play_backward()

        return super(ModuleGUI, self).eventFilter(QObject, QEvent)

    def play_forward(self):
        self.vi.viewer.play(500)

    def play_backward(self):
        self.vi.viewer.play(-500)

    def img_seq_home(self):
        self.vi.viewer.setCurrentIndex(0)
        self.vi.viewer.play(0)

    def img_seq_end(self):
        self.vi.viewer.setCurrentIndex(self.vi.viewer.getProcessedImage().shape[0] - 1)
        self.vi.viewer.play(0)

    def _list_widget_context_menu_requested(self, p):
        self.list_widget_context_menu.exec_(self.ui.listWidgetROIs.mapToGlobal(p))

    def slot_delete_roi_menu(self):
        """Delete the currently selected ROI"""
        if hasattr(self, 'manager'):
            if isinstance(self.manager, managers.ManagerCNMFROI):
                self.manager.update_idx_components(self.manager.roi_list.current_index)

            del self.manager.roi_list[self.manager.roi_list.current_index]

    def btnAddROI_context_menu_requested(self, p):
        self.btnAddROIMenu.exec_(self.ui.btnAddROI.mapToGlobal(p))

    def start_backend(self, type_str: str):
        """Choose backend, one of the Manager classes in the managers module."""

        if hasattr(self, 'manager'):
            del self.manager

        if type_str == 'Manual':
            self.start_manual_mode()
            self.disable_curve_options(True)
            self.ui.radioButton_curve_data.setChecked(True)
            return

        self.ui.btnAddROI.setDisabled(True)  # only used for manual ROIs
        self.disable_curve_options(False)  # enable dfof and spike options

        backend = f'Manager{type_str}'
        manager = getattr(managers, backend)

        self.manager = manager(self, self.ui, self.vi)
        self.vi.viewer.workEnv.roi_manager = self.manager

        self.ui.btnSwitchToManualMode.setEnabled(True)
        self.ui.radioButton_curve_data.setChecked(True)

        logger.info(f'ROI Manager backend set to: {backend}')

        return self.manager

    def start_manual_mode(self):
        """Start in manual mode. Creates a new back-end manager instance (Uses ManagerManual)"""
        logger.info('ROI back-end set to: ManagerManual')

        if hasattr(self, 'manager'):
            del self.manager
        self.manager = managers.ManagerManual(self, self.ui, self.vi)

        self.ui.btnAddROI.setEnabled(True)
        self.vi.viewer.workEnv.roi_manager = self.manager
        self.ui.btnSwitchToManualMode.setDisabled(True)

    def set_curveplot_datatype(self):
        if self.ui.radioButton_curve_data.isChecked():
            datatype = 'curve'
        elif self.ui.radioButton_dfof.isChecked():
            datatype = 'dfof'
        elif self.ui.radioButton_spikes.isChecked():
            datatype = 'spike'

        if isinstance(self.manager, managers.ManagerVolMultiCNMFROI):
            for roi in self.manager.roi_list:
                roi.curve_data_type = datatype
                if roi.visible:
                    roi.set_viewer_curveplot(datatype)

        else:
            for roi in self.manager.roi_list:
                roi.set_viewer_curveplot(datatype)

    def disable_curve_options(self, b):
        self.ui.radioButton_curve_data.setDisabled(b)
        self.ui.radioButton_dfof.setDisabled(b)
        self.ui.radioButton_spikes.setDisabled(b)

    def add_manual_roi(self, shape: str):
        """Add a manual ROI. Just calls ManagerManual.add_roi"""
        if self.vi.viewer.workEnv.isEmpty:
            QtWidgets.QMessageBox.warning(self,
                                          'Empty Work Environment',
                                          'You cannot add ROIs to an empty work environment')
            return
        self.manager.add_roi(shape)

    def add_roi_tags_list_text(self):
        pass

    def show_all_rois(self):
        pass

    def package_for_project(self) -> dict:
        """
        Gets all the ROI states so that they can be packaged along with
        the rest of the work environment to be saved as a project Sample
        """

        states = self.manager.get_all_states()
        return states

    def set_all_from_states(self, states: dict):
        """Set all the ROIs from a states dict. Instantiates the appropriate back-end Manager"""
        if 'roi_type' not in states.keys():
            raise TypeError('`roi_type` not specified in states dict')

        roi_class = states["roi_type"]

        if roi_class not in globals().keys():
            raise TypeError(f'Specified ROI type <{roi_class}> not found in available ROI types.')

        # All scatter types, CNMF(E), Suite2p, VolCNMF etc.
        if issubclass(globals()[roi_class], ScatterROI):
            self.start_backend(states['roi_type'])
            self.manager.restore_from_states(states)

        # Manual ROI classes
        elif issubclass(globals()[roi_class], ManualROI):
            self.start_manual_mode()
            self.manager.restore_from_states(states)

    def import_from_imagej(self):
        """Import ROIs from ImageJ zip file"""
        path = QtWidgets.QFileDialog.getOpenFileName(None, 'Import ImageJ ROIs', '', '(*.zip)')
        if path == '':
            return
        self.start_manual_mode()
        try:
            self.manager.import_from_imagej(path[0])
        except:
            QtWidgets.QMessageBox.warning(None, 'File open Error',
                                          'Could not open the chosen file.\n'
                                          'It might not be an ImageJ ROIs file' + traceback.format_exc())

    def set_spot_size(self, size: int):
        if hasattr(self.manager, 'set_spot_size'):
            self.manager.set_spot_size(size)

    def request_zlevel(self, roi_ix):
        if not hasattr(self, 'manager'):
            return

        if not isinstance(self.manager, managers.ManagerVolROI):
            return

        roi = self.manager.roi_list[roi_ix]

        if not roi.visible:
            self.vi.viewer.ui.verticalSliderZLevel.setValue(roi.zcenter)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tab_page.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_TabPage(object):
    def setupUi(self, TabPage):
        TabPage.setObjectName("TabPage")
        TabPage.resize(589, 371)
        self.gridLayout = QtWidgets.QGridLayout(TabPage)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(482, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 0, 1, 1)
        self.btnAddRow = QtWidgets.QPushButton(TabPage)
        self.btnAddRow.setObjectName("btnAddRow")
        self.gridLayout.addWidget(self.btnAddRow, 1, 3, 1, 1)
        self.scrollArea = QtWidgets.QScrollArea(TabPage)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 573, 315))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setObjectName("verticalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 4)
        self.comboBoxTimeUnits = QtWidgets.QComboBox(TabPage)
        self.comboBoxTimeUnits.setEnabled(False)
        self.comboBoxTimeUnits.setObjectName("comboBoxTimeUnits")
        self.comboBoxTimeUnits.addItem("")
        self.comboBoxTimeUnits.addItem("")
        self.gridLayout.addWidget(self.comboBoxTimeUnits, 1, 2, 1, 1)
        self.label = QtWidgets.QLabel(TabPage)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 1, 1, 1)

        self.retranslateUi(TabPage)
        QtCore.QMetaObject.connectSlotsByName(TabPage)

    def retranslateUi(self, TabPage):
        _translate = QtCore.QCoreApplication.translate
        TabPage.setWindowTitle(_translate("TabPage", "Form"))
        self.btnAddRow.setText(_translate("TabPage", "Add Row"))
        self.comboBoxTimeUnits.setItemText(0, _translate("TabPage", "frames"))
        self.comboBoxTimeUnits.setItemText(1, _translate("TabPage", "seconds"))
        self.label.setText(_translate("TabPage", "units: "))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 17 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from .dataframe_columns_widget import DataFrameColumnsWidget
from .column_widget import ColumnWidget
import pandas as pd
# from common import configurationt
from numpy import int64, float64
from copy import deepcopy
# from functools import partial
import traceback
from ...common import configuration, get_project_manager


MOD_NOT = '$NOT:'
MOD_STR = '$STR:'
MOD_STR_EQ = '$STR=:'
MOD_STR_NOT_EQ = '$STR!=:'
MOD_GT = '$>:'
MOD_LT = '$<:'
MOD_GE = '$>=:'
MOD_LE = '$<=:'


class TabAreaWidget(QtWidgets.QWidget):
    signal_new_tab_requested = QtCore.pyqtSignal(pd.DataFrame, list)
    signal_open_sample_in_viewer_requested = QtCore.pyqtSignal(str)

    def __init__(self, parent, tab_name: str, dataframe: pd.DataFrame, filter_history: list, is_root=False):
        QtWidgets.QWidget.__init__(self, parent)

        self.vertical_layout = QtWidgets.QVBoxLayout(self)
        self.columns_widget = DataFrameColumnsWidget(parent=self)
        self.vertical_layout.addWidget(self.columns_widget)

        self.tab_name = tab_name
        self.is_root = is_root
        self.filter_history = filter_history
        self.dataframe = dataframe

        self.columns = []

        # def initialize_gui(self):
        for column in self.dataframe.columns:
            if column in configuration.proj_cfg['EXCLUDE'].keys():
                continue

            c = ColumnWidget(parent=self.columns_widget,
                             tab_name=self.tab_name,
                             column_name=column,
                             is_root=self.is_root)

            c.signal_apply_clicked.connect(self.slot_filter_requested)
            if column == 'SampleID':
                c.signal_sample_id.connect(self.emit_open_sample_in_viewer_requested)
            self.columns.append(c)
            self.columns_widget.add_column(c)

        self.populate_tab()
        # self.columns_widget.show()

    @QtCore.pyqtSlot(str)
    def emit_open_sample_in_viewer_requested(self, sample_id: str):
        self.signal_open_sample_in_viewer_requested.emit(sample_id)

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    @dataframe.setter
    def dataframe(self, dataframe: pd.DataFrame):
        self._dataframe = dataframe

    def set_columns_empty(self):
        for column in self.columns:
            column.set_empty()

    @QtCore.pyqtSlot(dict)
    def slot_filter_requested(self, d: dict):
        column_widget = d['column_widget']

        dataframe = self.dataframe.copy()

        if (d['option'] is None) or (d['option'] == 'new'):
            try:
                dataframe, history = self._filter(column_widget, dataframe)
            except (ValueError, TypeError) as e:
                QtWidgets.QMessageBox.warning(self, 'Invalid entry', str(e))
                return

        elif (d['option'] == 'all') or (d['option'] == 'all_in_new'):
            history = []
            for column in self.columns:
                assert isinstance(column, ColumnWidget)
                if column.lineEdit.text() != '':
                    try:
                        dataframe, h = self._filter(column, dataframe)
                        history += h
                    except (ValueError, TypeError) as e:
                        QtWidgets.QMessageBox.warning(self, 'Invalid entry', str(e))
                        return
        filter_history = deepcopy(self.filter_history)
        filter_history += history
        if self.is_root or (d['option'] == 'all_in_new') or (d['option'] == 'new'):
            self.signal_new_tab_requested.emit(dataframe, filter_history)
            return

        else:
            self.dataframe = dataframe
            self.filter_history += filter_history
            self.populate_tab()

        get_project_manager().child_dataframes[self.tab_name]['dataframe'] = self.dataframe
        get_project_manager().child_dataframes[self.tab_name]['filter_history'] = self.filter_history

    def _filter(self, column_widget: ColumnWidget, dataframe: pd.DataFrame) -> tuple:
        filt = column_widget.lineEdit.text()

        if filt.startswith('$') \
                and not filt.startswith(('$NOT:',
                                         '$STR:',
                                         '$STR=:',
                                         '$STR!=:',
                                         '$>:',
                                         '$<:',
                                         '$>=:',
                                         '$<=:')):

            raise ValueError('Unrecognized modifer in column ' + str(column_widget.column_name))

        if column_widget.column_type is list:
            if filt.startswith(MOD_NOT):  # NOT
                filt = filt[len(MOD_NOT):].split('|')
                selection = ~dataframe[column_widget.column_name].apply(
                    lambda l: any([v in u for u in l for v in filt])  # True if any str in the `filt` list is a
                )                                                     # substring in the `l` list
                selection_str = \
                    f'~df["{column_widget.column_name}"].apply(' \
                            f'lambda l: any([v in u for u in l for v in {str(filt)}])'

            elif filt.startswith(MOD_STR):  # treat as str, same as not using this modifier for this case
                filt = filt[len(MOD_STR):].split('|')
                selection = dataframe[column_widget.column_name].apply(
                    lambda l: any([v in u for u in l for v in filt])
                )
                selection_str = \
                    f'df["{column_widget.column_name}"].apply(' \
                        f'lambda l: any([v in u for u in l for v in {str(filt)}])'

            elif filt.startswith(MOD_STR_EQ):  # exact match of str
                filt = filt[len(MOD_STR_EQ):].split('|')
                selection = dataframe[column_widget.column_name].apply(
                    lambda l: any([v == u for u in l for v in filt])  # True if any str in the `filt` list is
                )                                                     # an exact match of any strings in `l`
                selection_str = \
                    f'df["{column_widget.column_name}"].apply(' \
                        f'lambda l: any([v == u for u in l for v in {str(filt)}])'

            elif filt.startswith(MOD_STR_NOT_EQ):  # negation of exact match of str
                filt = filt[len(MOD_STR_NOT_EQ):].split('|')
                selection = ~dataframe[column_widget.column_name].apply(
                    lambda l: any([v == u for u in l for v in filt])
                )
                selection_str = \
                    f'~df["{column_widget.column_name}"].apply(' \
                        f'lambda l: any([v == u for u in l for v in {str(filt)}])'

            else:
                filt = filt.split('|')  # check if any of the input strings are a substring in the list
                selection = dataframe[column_widget.column_name].apply(
                    lambda l: any([v in u for u in l for v in filt])
                )
                selection_str = \
                    f'df["{column_widget.column_name}"].apply(' \
                        f'lambda l: any([v in u for u in l for v in {str(filt)}])'

        elif column_widget.column_type is str or \
                filt.startswith('$STR:') or \
                filt.startswith('$STR=:') or \
                filt.startswith('$STR!=:'):

            if filt.startswith('$NOT:'):
                filt = filt[5:]
                selection = ~dataframe[column_widget.column_name].str.contains(filt)
                selection_str = '~df["' + column_widget.column_name + '"].str.contains("' + filt + '")'

            elif filt.startswith('$STR:'):
                filt = filt[5:]
                selection = dataframe[column_widget.column_name].str.contains(filt)
                selection_str = 'df["' + column_widget.column_name + '"].str.contains("' + filt + '")'

            elif filt.startswith('$STR=:'):
                filt = filt[6:].split('|')
                selection = dataframe[column_widget.column_name].isin(filt)
                selection_str = 'df["' + column_widget.column_name + '"].isin(' + str(filt) + ')'

            elif filt.startswith('$STR!=:'):
                filt = filt[7:].split('|')
                selection = ~dataframe[column_widget.column_name].isin(filt)
                selection_str = '~df["' + column_widget.column_name + '"].isin(' + str(filt) + ')'

            else:
                selection = dataframe[column_widget.column_name].str.contains(filt)
                selection_str = 'df["' + column_widget.column_name + '"].str.contains("' + filt + '")'

        elif column_widget.column_type in [int, float, int64, float64]:
            filt = column_widget.lineEdit.text()

            if column_widget.column_type is int:
                t_func = 'int'
            elif column_widget.column_type is float:
                t_func = 'float'
            elif column_widget.column_type is int64:
                t_func = 'int64'
            elif column_widget.column_type is float64:
                t_func = 'float64'

            if filt.startswith('$NOT:'):
                filt = filt[5:].split('|')
                filt = list(map(column_widget.column_type, filt))
                selection = ~dataframe[column_widget.column_name].isin(filt)
                selection_str = '~df["' + column_widget.column_name + '"].isin(' + str(filt) + ')'

            elif filt.startswith('$') and '|' in filt:
                raise ValueError('Nonsensical filter\nYou cannot use ">" or "<" operators in combination with "|"')

            elif filt.startswith('$>:'):
                filt = column_widget.column_type(filt[3:])
                selection = dataframe[column_widget.column_name] > column_widget.column_type(filt)
                selection_str = 'df["' + column_widget.column_name + '"] > ' + t_func + '(' + str(filt) + ')'

            elif filt.startswith('$<:'):
                filt = column_widget.column_type(filt[3:])
                selection = dataframe[column_widget.column_name] < column_widget.column_type(filt)
                selection_str = 'df["' + column_widget.column_name + '"] < ' + t_func + '(' + str(filt) + ')'

            elif filt.startswith('$>=:'):
                filt = column_widget.column_type(filt[4:])
                selection = dataframe[column_widget.column_name] >= column_widget.column_type(filt)
                selection_str = 'df["' + column_widget.column_name + '"] >= ' + t_func + '(' + str(filt) + ')'

            elif filt.startswith('$<=:'):
                filt = column_widget.column_type(filt[4:])
                selection = dataframe[column_widget.column_name] <= column_widget.column_type(filt)
                selection_str = 'df["' + column_widget.column_name + '"] <= ' + t_func + '(' + str(filt) + ')'

            else:
                filt = filt.split('|')
                filt = list(map(column_widget.column_type, filt))
                selection = dataframe[column_widget.column_name].isin(filt)
                selection_str = 'df["' + column_widget.column_name + '"].isin(' + str(filt) + ')'

        else:
            raise TypeError('Unsupported type for filtering, '
                            'can only support str, list, int, float, '
                            'numpy.int64 and numpy.float64')

        filtered_df = dataframe[selection]
        filter_history = 'df[' + selection_str + ']'

        return filtered_df, [filter_history]

    def populate_tab(self):
        if self.dataframe.size < 1:
            self.set_columns_empty()
            return

        for column in self.columns:
            try:
                self._populate_column(column, self.dataframe[column.column_name])
            except Exception:
                raise ValueError(f"Cannot open project due to an issue with the following column: `{column.column_name}`\n")

    def _populate_column(self, column: ColumnWidget, series: pd.Series):
        column.series = series



# -*- coding: utf-8 -*-
import traceback
from PyQt5 import QtWidgets
from ....plotting.widgets.peak_editor import peak_editor
from .common import *
from ....analysis import Transmission
from scipy import signal
from scipy import fftpack
import pandas as pd
from ....analysis.compute_peak_features import ComputePeakFeatures
from ....common.configuration import HAS_TSLEARN
if HAS_TSLEARN:
    from tslearn.preprocessing import TimeSeriesScalerMeanVariance


class ButterWorth(CtrlNode):
    """Butterworth Filter"""
    nodeName = 'ButterWorth'
    uiTemplate = [('data_column', 'combo', {}),
                  ('order', 'intSpin', {'min': 1, 'max': 100, 'step': 1, 'value': 2}),
                  ('freq_divisor', 'doubleSpin', {'min': 0.01, 'max': 100.00, 'step': 0.05, 'value': 2.00}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    output_column = '_BUTTERWORTH'

    def _func(self, x: np.ndarray, meta: dict):
        N = self.ctrls['order'].value()
        freq = 1 / meta['fps']

        self.Wn = freq/self.freq_divisor

        b, a = signal.butter(N, self.Wn)
        sig = signal.filtfilt(b, a, x)

        return pd.Series({self.output_column: sig})

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()
        if self.ctrls['Apply'] is False:
            return

        self.t = transmission.copy()

        self.order = self.ctrls['order'].value()
        self.freq_divisor = self.ctrls['freq_divisor'].value()

        # try:
        tqdm.pandas()
        self.t.df[self.output_column] = self.t.df.progress_apply(lambda x: self._func(x[self.data_column], x['meta']), axis=1)
        # except KeyError as e:
        #     raise KeyError(str(e))

        params = {'data_column': self.data_column,
                  'order': self.order,
                  'Wn': self.Wn,
                  'freq_divider': self.freq_divisor,
                  'units': self.t.last_unit
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='butterworth', parameters=params)
        self.t.last_output = self.output_column

        return self.t


class SavitzkyGolay(CtrlNode):  # Savitzky-Golay filter for example
    """Savitzky-Golay filter."""
    nodeName = 'Savitzky_Golay'
    uiTemplate = [('data_column', 'combo', {}),
                  ('window_length', 'intSpin', {'min': 3, 'max': 999, 'value': 3, 'step': 2}),
                  ('polyorder', 'intSpin', {'min': 1, 'max': 998, 'value': 1, 'step': 1}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        w = self.ctrls['window_length'].value()
        p = self.ctrls['polyorder'].value()

        if p > w:
            raise ValueError('Invalid value! polyorder MUST be less than window_length')

        if w % 2 == 0:
            raise ValueError('Invalid value! window_length MUST be an odd number!')

        output_column = '_SAVITZKY_GOLAY'

        tqdm.pandas()
        self.t.df[output_column] = self.t.df[self.data_column].progress_apply(signal.savgol_filter, window_length=w, polyorder=p)

        params = {'data_column': self.data_column,
                  'window_length': w,
                  'polyorder': p,
                  'units': self.t.last_unit
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='savitzky_golay', parameters=params)
        self.t.last_output = output_column

        return self.t


class PowerSpectralDensity(CtrlNode):
    """Return the Power Spectral Density of a curve."""
    nodeName = 'PowSpecDens'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        output_column = '_POWER_SPECTRAL_DENSITY'

        tqdm.pandas()
        self.t.df[output_column] = self.t.df[self.data_column].progress_apply(self._func)

        params = {'data_column': self.data_column}
        self.t.history_trace.add_operation(data_block_id='all', operation='power_spectral_density', parameters=params)
        self.t.last_output = output_column

        return self.t

    def _func(self, curve):
        f, p = signal.periodogram(curve)
        return p


class Resample(CtrlNode):
    """
    Resample 1D data, uses scipy.signal.resample. "Rs" is the new sampling rate in "Tu" units of time.
    If "Tu" = 1, then Rs is the new sampling rate in Hertz.
    """

    nodeName = 'Resample'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Rs', 'intSpin', {'min': 1, 'max': 9999, 'value': 10, 'step': 5}),
                  ('Tu', 'intSpin', {'min': 1, 'max': 9999, 'value': 1, 'step': 1}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        self.Rs = self.ctrls['Rs'].value()
        self.Tu = self.ctrls['Tu'].value()
        self.new_rate = self.Rs / self.Tu

        output_column = '_RESAMPLE'

        tqdm.pandas()
        self.t.df[output_column] = self.t.df.progress_apply(self._func, axis=1)

        params = {'data_column': self.data_column,
                  'output_rate': self.new_rate,
                  'units': self.t.last_unit
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='resample', parameters=params)
        self.t.last_output = output_column

        return self.t

    def _func(self, row: pd.Series):
        Nf = row['meta']['fps']
        if Nf == 0:
            raise ValueError('Framerate not set for SampleID: ' + row['SampleID']
                             + '. You must set a framerate for this SampleID to continue')
        Ns = row[self.data_column].shape[0]

        Rn = int((Ns / Nf) * (self.new_rate))

        return signal.resample(row[self.data_column], Rn)


class ScalerMeanVariance(CtrlNode):
    """Scaler for time series. Scales time series so that their mean (resp. standard deviation) in each dimension is mu (resp. std).\n
    See https://tslearn.readthedocs.io/en/latest/gen_modules/preprocessing/tslearn.preprocessing.TimeSeriesScalerMeanVariance.html#tslearn.preprocessing.TimeSeriesScalerMeanVariance"""
    nodeName = 'ScalerMeanVar'
    uiTemplate = [('data_column', 'combo', {}),
                   ('mu', 'doubleSpin', {'value': 0.0, 'step': 0.1, 'toolTip': 'Mean of the output time series'}),
                   ('std', 'doubleSpin', {'value': 1.0, 'step': 1.0, 'toolTip': 'Standard deviation of the output time series'}),
                   ('Apply', 'check', {'checked': False, 'applyBox': True})]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if not self.ctrls['Apply'].isChecked():
            return

        self.t = transmission.copy()

        mu = self.ctrls['mu'].value()
        std = self.ctrls['std'].value()

        params = {'data_column': self.data_column,
                  'mu': mu,
                  'std': std,
                  'units': self.t.last_unit
                  }

        output_column = '_SCALER_MEAN_VARIANCE'

        self.t.df[output_column] = self.t.df[self.data_column].apply(lambda a: TimeSeriesScalerMeanVariance(mu=mu, std=std).fit_transform(a)[:, :, 0])
        self.t.history_trace.add_operation(data_block_id='all', operation='scaler_mean_variance', parameters=params)
        self.t.last_output = output_column

        return self.t


class Normalize(CtrlNode):
    """Normalize a column containing 1-D arrays such that values in each array are normalized between 0 and 1\n
    Output Column -> Input Column"""

    nodeName = 'Normalize'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()
        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        output_column = '_NORMALIZE'

        self.t.df[output_column] = self.t.df[self.data_column].apply(lambda a: ((a - np.min(a)) / (np.max(a - np.min(a)))))

        params = {'data_column': self.data_column,
                  'units': self.t.last_unit
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='normalize', parameters=params)
        self.t.last_output = output_column

        return self.t


class RFFT(CtrlNode):
    """
    Uses fftpack.rfft, 'Discrete Fourier transform of a real sequence.\n
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.fftpack.rfft.html#scipy.fftpack.rfft
    """
    nodeName = 'RFFT'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        # TODO: Replace this code block with the DataTypes.get_framerate() method
        ##############################################################################################################
        sampling_rates = []

        for db in transmission.history_trace.data_blocks:
            if transmission.history_trace.check_operation_exists(db, 'resample'):
                sampling_rates.append(transmission.history_trace.get_operation_params(db, 'resample')['output_rate'])
            else:
                r = pd.DataFrame(transmission.get_data_block_dataframe(db).meta.to_list())['fps'].unique()
                # if rates.size > 1:
                #     raise ValueError("Sampling rates for the data do not match")
                # else:
                sampling_rates.append(r)

        rates = np.hstack([sampling_rates])

        if np.ptp(rates) > 0.1:
            raise ValueError("Sampling rates of the data differ by greater than the tolerance of 0.1 Hz")

        framerate = int(np.mean(sampling_rates))
        ##############################################################################################################

        array_size = transmission.df[self.data_column].apply(lambda a: a.size).unique()

        if array_size.size > 1:
            raise ValueError("Size of all arrays in data column must match exactly.")

        array_size = array_size[0]

        freqs = fftpack.rfftfreq(array_size) * framerate

        self.t = transmission.copy()

        output_column = '_RFFT'

        self.t.df[output_column] = self.t.df[self.data_column].apply(fftpack.rfft)

        self.t.last_unit = 'frequency'

        params = {'data_column':    self.data_column,
                  'frequencies':    freqs.tolist(),
                  'sampling_rate':  framerate,
                  'nyquist_frequency': float(freqs.max()),
                  'units': 'frequency'
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='rfft', parameters=params)
        self.t.last_output = output_column

        return self.t


class iRFFT(CtrlNode):
    """
    Uses fftpack.irfft, 'Return inverse discrete Fourier transform of real sequence.'\n
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.fftpack.irfft.html#scipy.fftpack.irfft\n
    Input must have an _RFFT column from the RFFT node.\n
    """
    nodeName = 'iRFFT'
    uiTemplate = [('Apply', 'check', {'checked': False, 'applyBox': True})]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()
        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        output_column = '_IRFFT'

        self.t.df[output_column] = self.t.df['_RFFT'].apply(fftpack.irfft)
        self.t.last_unit = 'time'

        params = {'data_column': '_RFFT',
                  'units': 'time'
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='irfft', parameters=params)
        self.t.last_output = output_column

        return self.t


class PeakDetect(CtrlNode):
    """Detect peaks & bases by finding local maxima & minima. Use this after the Derivative Filter"""
    nodeName = 'PeakDetect'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Fictional_Bases', 'check', {'checked': True}),
                  ('Edit', 'button', {'text': 'Open GUI'}),
                  ('SlopeThr', 'doubleSpin', {'min': -100.00, 'max': 1.0, 'step': 0.010}),
                  ('AmplThrAbs', 'doubleSpin', {'min': 0.00, 'max': 9999.00, 'step': 0.05}),
                  ('AmplThrRel', 'doubleSpin', {'min': 0.00, 'max': 9999.00, 'step': 0.05}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def __init__(self, name, **kwargs):
        CtrlNode.__init__(self, name, terminals={'Derivative': {'io': 'in'},
                                                 'Normalized': {'io': 'in'},
                                                 'Curve': {'io': 'in'},
                                                 'PB_Input': {'io': 'in'},
                                                 'Out': {'io': 'out', 'bypass': 'Curve'}}, **kwargs)
        self.data_modified = False
        self.editor_output = False
        self.pbw = None
        self.ctrls['Edit'].clicked.connect(self._peak_editor)
        self.t = None
        self.params = {}

    def _get_zero_crossings(self, d1: np.ndarray, sig: np.ndarray, norm_sig: np.ndarray, fictional_bases: bool = False) -> pd.DataFrame:
        """
        Find the peaks and bases of the signal by finding zero crossing in the first derivative of the filtered signal
        :param dsig: The first derivative of the signal
        :return: DataFrame, all zero crossing events in one column, another column denotes it as a peak or base.
        """
        # Get array of all sign switches
        sc = np.diff(np.sign(d1))

        peaks_raw = np.where(sc < 0)[0]
        bases = np.where(sc > 0)[0]

        # Remove all peaks where amplitude is below the specified threshold
        peak_yvals = np.take(norm_sig, peaks_raw)
        # print('peak_yvals: ' + str(peak_yvals))
        ix_below_ampl_thr = np.where(peak_yvals < self.ctrls['AmplThrAbs'].value())
        # print('ix_below_ampl_thr: ' + str(ix_below_ampl_thr))
        peaks_ampl_thr = np.delete(peaks_raw, ix_below_ampl_thr)

        s2 = np.gradient(d1)
        # Remove all peaks where the 2nd derivative is below a certain threshold
        peak_d2 = np.take(s2, peaks_ampl_thr)
        ix_below_slope_thr = np.where(peak_d2 > self.ctrls['SlopeThr'].value())
        # print('peak_d2: ' + str(peak_d2))
        # print('ix_below_slope_thr: ' + str(ix_below_slope_thr))
        peaks = np.delete(peaks_ampl_thr, ix_below_slope_thr)

        ## TODO; DEBATE ABOUT HOW TO PROPERLY DEAL WITH TRACES THAT HAVE NO PEAKS
        if peaks.size == 0:
            # abort = True
            # peaks = np.array([1])
            # bases = np.array([0, 2])
            return pd.DataFrame()
        # else:
        #     self.row_ix += 1
        #     return pd.DataFrame()

        # Add bases to beginning and end of sequence if first or last peak is lonely
        if fictional_bases:
            if bases.size == 0:
                bases = np.array([0, sc.size])
            else:
                if bases[0] > peaks[0]:
                    bases = np.insert(bases, 0, 0)

                if bases[-1] < peaks[-1]:
                    bases = np.insert(bases, -1, sc.size)

        # Construct peak & base columns on dataframe
        peaks_df = pd.DataFrame()
        peaks_df['event'] = peaks
        peaks_df['label'] = 'peak'

        bases_df = pd.DataFrame()
        bases_df['event'] = bases
        bases_df['label'] = 'base'

        peaks_bases_df = pd.concat([peaks_df, bases_df])
        peaks_bases_df = peaks_bases_df.sort_values('event')
        peaks_bases_df.reset_index(drop=True, inplace=True)

        # if abort:
        #     self.row_ix += 1
        #     warn(f'No peaks detected at index: {self.row_ix}')
        #     return peaks_bases_df

        # peaks_bases_df['peak'] = peaks_bases_df['label'] == 'peak'
        # peaks_bases_df['base'] = peaks_bases_df['label'] == 'base'

        # Set the peaks at the index of the local maxima of the raw curve instead of the maxima inferred
        # from the derivative
        # Also remove peaks which are lower than the relative amplitude threshold
        rows_drop = []
        for ix, r in peaks_bases_df.iterrows():
            if r['label'] == 'peak' and ix > 0:
                if peaks_bases_df.iloc[ix - 1]['label'] == 'base' and peaks_bases_df.iloc[ix + 1]['label'] == 'base':
                    ix_left_base = peaks_bases_df.iloc[ix - 1]['event']
                    ix_right_base = peaks_bases_df.iloc[ix + 1]['event']

                    #  Adjust the xval of the curve by finding the absolute maxima of this section of the raw curve,
                    # flanked by the bases of the peak
                    peak_revised = np.where(sig == np.max(
                        np.take(sig, np.arange(ix_left_base, ix_right_base))))[0][0]

                    # Get rising and falling amplitudes
                    rise_ampl = sig[peak_revised] - sig[ix_left_base]
                    fall_ampl = sig[peak_revised] - sig[ix_right_base]

                    # Check if above relative amplitude threshold
                    if ((rise_ampl + fall_ampl) / 2) > self.ctrls['AmplThrRel'].value():
                        peaks_bases_df.set_value(ix, 'event', peak_revised)
                    else:
                        rows_drop.append(ix)

        peaks_bases_df = peaks_bases_df.drop(peaks_bases_df.index[rows_drop])
        peaks_bases_df = peaks_bases_df.reset_index(drop=True)

        # remove bases that aren't around any peak
        for ix, r in peaks_bases_df.iterrows():
            if r['label'] == 'base' and 1 < ix < (peaks_bases_df.index.size - 1):
                if peaks_bases_df.iloc[ix - 1]['label'] != 'peak' and peaks_bases_df.iloc[ix + 1]['label'] != 'peak':
                    rows_drop.append(ix)

        # Weird behavior dealing with bases at the end of a curve
        try:
            peaks_bases_df = peaks_bases_df.drop(peaks_bases_df.index[rows_drop])
            peaks_bases_df = peaks_bases_df.reset_index(drop=True)
        except:
            pass

        self.row_ix += 1
        # print(peaks_bases_df)
        return peaks_bases_df

    def process(self, display=True, **kwargs):
        out = self.processData(**kwargs)
        return {'Out': out}

    def processData(self, **inputs):
        if self.editor_output:
            return self.t

        pb_input = inputs['PB_Input']
        if isinstance(pb_input, Transmission):
            self.t = pb_input
            self.t.df.reset_index(drop=True, inplace=True)
            if self.pbw is None:
                self._peak_editor()
            else:
                self.pbw.update_transmission(self.t, self.t)
            self.pbw.btnDisconnectFromFlowchart.setDisabled(True)
            return self.t

        columns = inputs['Curve'].df.columns.to_list()
        self.ctrls['data_column'].setItems(organize_dataframe_columns(columns)[0])

        if self.data_modified is True:
            if QtWidgets.QMessageBox.question(None, 'Discard peak edits?',
                                              'You have made modifications to peak data passing '
                                              'through this node! Would you like to discard all '
                                              'changes and load the newly transmitted data?',
                                              QtWidgets.QMessageBox.Yes,
                                              QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
                return self.t
        self.data_modified = False

        if not self.ctrls['Apply'].isChecked():
            return
        # if inputs['Derivative'] is None:
        #     raise Exception('No incoming Derivative transmission. '
        #                     'You must input at least a derivative')
        #
        # self.t = inputs['Derivative'].copy()
        #
        # if inputs['Curve'] is not None:
        #     if inputs['Derivative'].df.index.size != inputs['Curve'].df.index.size:
        #         QtWidgets.QMessageBox.warning(None, 'ValueError!', 'Input diemensions of Derivative and Curve transmissions'
        #                                                        ' MUST match!')
        #         raise ValueError('Input diemensions of Derivative and Curve transmissions MUST match!')

        if inputs['Derivative'] is None:
            raise KeyError('No incoming Derivative transmission. '
                           'You must input both a curve and its derivative '
                           'You must input at least a derivative')

        if inputs['Curve'] is None:
            raise KeyError('No incoming Curve transmission.'
                           ' You must input both a curve and its derivative')

        if inputs['Derivative'].df.index.size != inputs['Curve'].df.index.size:
            raise ValueError('Input diemensions of Derivative and Curve transmissions MUST match!')

        t_d1 = inputs['Derivative'].copy()
        assert isinstance(t_d1, Transmission)
        d1 = t_d1.df['_DERIVATIVE']

        t_norm = inputs['Normalized'].copy()
        assert isinstance(t_norm, Transmission)
        norm_series = t_norm.df[t_norm.last_output]

        self.t = inputs['Curve'].copy()
        self.t.df['_DERIVATIVE'] = d1
        self.t.df['_NORM_PD'] = norm_series
        data_column = self.ctrls['data_column'].currentText()
        # self.t.df['raw_curve'] = self.t.df[data_column]

        assert isinstance(self.t, Transmission)

        # self.t.data_column['peaks_bases'] = 'peaks_bases'
        fb = self.ctrls['Fictional_Bases'].isChecked()
        self.row_ix = 0

        tqdm.pandas()
        self.t.df['peaks_bases'] = self.t.df.progress_apply(lambda r: self._get_zero_crossings(r['_DERIVATIVE'], r[data_column], r['_NORM_PD'], fb), axis=1)

        self.t.df['curve'] = self.t.df[data_column]
        self.t.df.drop(columns=['_NORM_PD'], inplace=True)
        self.t.df.reset_index(inplace=True, drop=True)
        # self.t.df.drop(columns=['raw_curve'])

        if self.pbw is not None:
            # self.pbw.curve_column = data_column
            self.pbw.update_transmission(self.t, self.t)

        params = {'data_column': data_column,
                  'SlopeThr': self.ctrls['SlopeThr'].value(),
                  'AmplThrRel': self.ctrls['AmplThrRel'].value(),
                  'AmplThrAbs': self.ctrls['AmplThrAbs'].value(),
                  'derivative_input_history_trace': t_d1.history_trace.get_all_data_blocks_history(),
                  'normalized_input_history_trace': t_norm.history_trace.get_all_data_blocks_history(),
                  'units': self.t.last_unit
                  }

        self.params = params

        self.t.history_trace.add_operation('all', operation='peak_detect', parameters=params)

        return self.t

    def _set_editor_output(self, edited_transmission: Transmission):
        self.t = edited_transmission
        self.t.history_trace.add_operation('all', operation='peak_detect', parameters=self.params)
        self.editor_output = True
        self.data_modified = True
        # self.pbw.close()
        # self.editor_output = True
        # self.t = self.pbw.getData()
        # self.pbw = None
        # self.data_modified = True
        self.changed()

    def _peak_editor(self):
        if self.pbw is None:
            self.pbw = peak_editor.PeakEditorWindow(self.t, self.t)
            self.pbw.sig_send_data.connect(self._set_editor_output)
            self.pbw.sig_reconnect_flowchart.connect(self.changed)
            self.pbw.setWindowTitle(self.name())

        self.pbw.show()


class PeakFeatures(CtrlNode):
    """Extract peak features after peak detection"""
    nodeName = 'PeakFeatures'
    uiTemplate = [('data_column', 'combo', {}),
                  # ('Compute', 'button', {'text': 'compute'}),
                  ('Apply', 'check', {'applyBox': True, 'checked': False})
                  ]

    def __init__(self, *args, **kwargs):
        super(PeakFeatures, self).__init__(*args, **kwargs)
        self.peak_results = None

    def processData(self, transmission: Transmission):
        dcols = organize_dataframe_columns(transmission.df.columns)[0]
        self.ctrls['data_column'].setItems(dcols)

        if not self.apply_checked():
            return

        self.computer = ComputePeakFeatures()

        self.t = self.computer.compute(transmission=transmission.copy(), data_column=self.data_column)

        params = {'data_column': self.data_column}
        self.t.history_trace.add_operation('all', 'peak-features', params)

        # Data units cannot be relied on anymore since any user specifid data column
        # can be used for computing peak features
        self.t.last_output = None

        return self.t


class SigmaMAD(CtrlNode):
    nodeName = 'SigmaMAD'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Apply', 'check', {'applyBox': True, 'checked': True})
                  ]

    def processData(self, transmission: Transmission):
        dcols = organize_dataframe_columns(transmission.df.columns)[0]
        self.ctrls['data_column'].setItems(dcols)

        if not self.apply_checked():
            return

        self.t = transmission.copy()

        output_column = '_' + self.__class__.__name__.upper()
        params = {'data_column': self.data_column}

        self.t.df[output_column] = self.t.df[self.data_column].apply(lambda x: np.median(np.abs(x - np.median(x))))
        self.t.history_trace.add_operation('all', self.__class__.__name__.lower(), params)

        return self.t



# -*- coding: utf-8 -*-
from ...Qt import QtGui, QtCore, QtWidgets
from spyder.widgets.variableexplorer.objecteditor import oedit
from .common import *
import traceback
from ....analysis import Transmission
from ....analysis.history_widget import HistoryTreeWidget
from ....common import get_project_manager
from ....common.qdialogs import *
import os
import pickle
from glob import glob
from ....common.configuration import HAS_TSLEARN
if HAS_TSLEARN:
    from tslearn.preprocessing import TimeSeriesScalerMinMax


class LoadProjDF(CtrlNode):
    """Load raw project DataFrames as Transmission"""
    nodeName = 'Load_Proj_DF'
    uiTemplate = [('DF_Name', 'combo'),
                  ('Update', 'button', {'text': 'Update', 'toolTip': 'When clicked this node will update'
                                                                     ' from the project DataFrame'}),
                  ('Apply', 'check', {'applyBox': True, 'checked': False}),
                  ('PinDF', 'check', {'text': 'Yes', 'toolTip': 'Pin the DataFrame to the flowchart, this way\n'
                                                                'you can open another project and still propogate\n'
                                                                'the data from this node.'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'Out': {'io': 'out'}})
        self._loadNode = True
        self.t = None
        child_df_names = ['root'] + list(get_project_manager().child_dataframes.keys())
        self.ctrls['DF_Name'].addItems(child_df_names)
        self.ctrls['Update'].clicked.connect(self.changed)
        # print('Node Refs:')
        # print(configuration.df_refs)

    def process(self):
        if self.ctrls['Apply'].isChecked() is False:
            return self.t

        # print('#######Weak Refs Dict########')
        # print(configuration.df_refs)

        if self.ctrls['PinDF'].isEnabled():
            if self.ctrls['PinDF'].isChecked():
                # self.t = self.t.copy()
                self.ctrls['PinDF'].setDisabled(True)
                self.ctrls['Update'].setDisabled(True)
                return {'Out': self.t}

            if self.ctrls['DF_Name'].currentText() == '':
                return
            child_df_name = self.ctrls['DF_Name'].currentText()
            if child_df_name == 'root':
                df = get_project_manager().dataframe
                filter_history = None
            else:
                df = get_project_manager().child_dataframes[child_df_name]['dataframe']
                filter_history = get_project_manager().child_dataframes[child_df_name]['filter_history']
            proj_path = get_project_manager().root_dir
            # print('*****************config df ref hex ID:*****************')
            # print(hex(id(df)))
            self.t = Transmission.from_proj(proj_path, df, sub_dataframe_name=child_df_name,
                                            dataframe_filter_history={'dataframe_filter_history': filter_history})

            # print('Tranmission dataframe hexID:')
            # print(hex(id(self.t.df)))

        return {'Out': self.t}


class LoadFile(CtrlNode):
    """Load Transmission data object from pickled file"""
    nodeName = 'LoadFile'
    uiTemplate = [('load_trn', 'button', {'text': 'Open .trn File'}),
                  ('proj_trns', 'combo', {}),
                  ('fname', 'label', {'text': ''}),
                  ('proj_path', 'button', {'text': 'Project Path'}),
                  ('proj_path_label', 'label', {'text': ''})
                  ]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'Out': {'io': 'out'}})
        self.ctrls['load_trn'].clicked.connect(lambda: self.load_file())
        self.ctrls['proj_path'].clicked.connect(self.dir_dialog_proj_path)

        self.t = None
        self._loadNode = True

        proj_path = get_project_manager().root_dir
        
        if proj_path is not None:
            trns = glob(os.path.join(proj_path, 'trns', '*.trn'))
            trns_names = [""] + [os.path.basename(f) for f in trns]
            self.ctrls['proj_trns'].setItems(trns_names)
            self.ctrls['proj_trns'].currentTextChanged.connect(lambda fname: self.load_file(os.path.join(get_project_manager().root_dir, 'trns', fname) if fname else False, qdialog=False))

    # def file_dialog_trn_file(self):
    #     path = QtWidgets.QFileDialog.getOpenFileName(None, 'Import Transmission object', '', '(*.trn *.ptrn)')
    #     if path == '':
    #         return
    #     self.load_file(path[0])

    @use_open_file_dialog('Load Transmission file', None, ['*.trn', '*.ptrn'])
    def load_file(self, path: str, qdialog=True):
        if not path:
            return
        try:
            self.t = Transmission.from_hdf5(path)
        except:
            QtWidgets.QMessageBox.warning(None, 'File open Error!', 'Could not open the chosen file.\n' + traceback.format_exc())
            return

        self.ctrls['fname'].setText(os.path.basename(path))

        proj_path = get_project_manager().root_dir
        if proj_path is not None:
            self._set_proj_path(proj_path)

        # print(self.transmission)
        # self.update()
        self.changed()

    def _set_proj_path(self, path: str):
        self.ctrls['proj_path_label'].setText(os.path.basename(path))
        self.t.set_proj_path(path)

    def dir_dialog_proj_path(self):
        path = QtWidgets.QFileDialog.getExistingDirectory(None, 'Select Project Folder')

        if path == '':
            return

        try:
            self._set_proj_path(path)
            self.changed()
        except (FileNotFoundError, NotADirectoryError) as e:
            QtWidgets.QMessageBox.warning(None, 'Invalid Project Folder', 'This is not a valid Mesmerize project\n' + e)
            return

    def process(self):
        return {'Out': self.t}


class Save(CtrlNode):
    """Save Transmission data object"""
    nodeName = 'Save'
    uiTemplate = [('saveBtn', 'button', {'text': 'OpenFileDialog'}),
                  ('path', 'label', {'text' : ''}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def __init__(self, name):
        # super(Save, self).__init__(name, terminals={'data': {'io': 'in'}})
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in'}})
        self._bypass = False
        self.bypassButton = None
        self.ctrls['saveBtn'].clicked.connect(self._fileDialog)
        self._saveNode = True
        self.saveValue = None

    def process(self, In, display=True):
        if In is not None:
            self._save(In)
        else:
            raise Exception('No incoming transmission to save!')

    def _fileDialog(self):
        path = QtWidgets.QFileDialog.getSaveFileName(None, 'Save Transmission as', '', '(*.trn)')
        if path == '':
            return
        if path[0].endswith('.trn'):
            path = path[0]
        else:
            path = path[0] + '.trn'

        self.ctrls['path'].setText(path)

    def _save(self, transmission):
        # self.ctrls['saveBtn'].clicked.connect(self._fileDialog)
        if self.ctrls['Apply'].isChecked is False:
            return

        if self.ctrls['path'].text() != '':
            try:
                transmission.to_hdf5(self.ctrls['path'].text())
            except:
                QtWidgets.QMessageBox.warning(None, 'File save error', 'Could not save the transmission to file.\n' + traceback.format_exc())


class Merge(CtrlNode):
    """Merge transmissions"""
    nodeName = 'Merge'
    uiTemplate = [('no controls', 'label')]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}, 'Out': {'io': 'out'}})

    def process(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)

        self.t = Transmission.merge(self.transmissions_list)

        return {'Out': self.t}


class ViewTransmission(CtrlNode):
    """View transmission using the spyder object editor"""
    nodeName = 'ViewData'
    uiTemplate = [('no controls', 'label')]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in'}})
        # self.edited_transmission = None

    def processData(self, transmission: Transmission):
        self.t = transmission.copy()
        oedit({'dataframe': self.t.df, 'history_trace': self.t.history_trace.history})
        # if self.edited is not None:
        #     self.edited.add_operation('all', 'object_editor', {})
        #     return self.edited

        # return self.t


class DropNa(CtrlNode):
    """Drop NaNs from the DataFrame"""
    nodeName = 'DropNaNs'
    uiTemplate = [('axis', 'combo', {'values': ['row', 'columns'], 'toolTip': 'Choose to drop NaNs from according to all '
                                                                              'rows, columns, or a specific column'}),
                  ('how', 'combo', {'values': ['any', 'all'], 'toolTip': 'any: drop from chosen axis if any element is NaN\n'
                                                                         'all: drop from chosen axis if all elements are NaN'}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})]

    def __init__(self, *args, **kwargs):
        super(DropNa, self).__init__(*args, **kwargs)

    def processData(self, transmission: Transmission):
        items = ['row', 'columns'] + ['---------'] + transmission.df.columns.to_list()
        self.ctrls['axis'].setItems(items)

        if not self.ctrls['Apply'].isChecked():
            return

        self.t = transmission.copy()

        axis = self.ctrls['axis'].currentText()

        if self.ctrls['axis'].currentIndex() < 2:
            if axis == 'row':
                axis = 0
            elif axis == 'columns':
                axis = 1

            how = self.ctrls['how'].currentText()
            self.t.df.dropna(axis=axis, how=how, inplace=True)
        else:
            how = None
            self.t.df = self.t.df[~self.t.df[axis].isna()]

        self.t.history_trace.add_operation('all', 'dropna', parameters={'axis': axis, 'how': how})

        return self.t


class ViewHistory(CtrlNode):
    """View History Trace of the input Transmission"""
    nodeName = 'ViewHistory'
    uiTemplate = [('Show', 'button')]

    def __init__(self, *args, **kwargs):
        super(ViewHistory, self).__init__(*args, **kwargs)
        self.history_widget = HistoryTreeWidget()
        self.ctrls['Show'].clicked.connect(self.history_widget.show)

    def processData(self, transmission: Transmission):
        self.history_widget.fill_widget(transmission.history_trace)


class iloc(CtrlNode):
    """Pass only one or multiple DataFrame Indices"""
    nodeName = 'iloc'
    uiTemplate = [('Index', 'intSpin', {'min': 0, 'step': 1, 'value': 0})
                  # ('Indices', 'lineEdit', {'text': '0', 'toolTip': 'Index numbers separated by commas'})
                  ]

    def processData(self, transmission):
        self.ctrls['Index'].setMaximum(transmission.df.index.size - 1)
        # self.ctrls['Index'].valueChanged.connect(
        #     partial(self.ctrls['Indices'].setText, str(self.ctrls['Index'].value())))

        # indices = [int(ix.strip()) for ix in self.ctrls['Indices'].text().split(',')]
        i = self.ctrls['Index'].value()
        self.t = transmission.copy()
        self.t.df = self.t.df.iloc[i]
        # self.t.src.append({'iloc': {'index': i}})
        return self.t


class SpliceArrays(CtrlNode):
    """Splice 1-D numpy arrays in a particular column."""
    nodeName = 'SpliceArrays'
    uiTemplate = [('Apply', 'check', {'checked': True, 'applyBox': True}),
                  ('data_column', 'combo', {}),
                  ('indices', 'lineEdit', {'text': '', 'placeHolder': 'start_ix:end_ix'})]
    output_column = '_SPLICE_ARRAYS'

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()
        indices = self.ctrls['indices'].text()

        if indices == '':
            return
        if ':' not in indices:
            return
        else:
            indices = indices.split(':')

        start_ix = int(indices[0])
        end_ix = int(indices[1])

        data_column = self.ctrls['data_column'].currentText()


        self.t.df[self.output_column] = self.t.df[data_column].apply(lambda a: a[start_ix:end_ix])

        params = {'data_column': data_column,
                  'start_ix': start_ix,
                  'end_ix': end_ix,
                  'units': self.t.last_unit
                  }

        self.t.history_trace.add_operation(data_block_id='all', operation='splice_arrays', parameters=params)
        self.t.last_output = self.output_column

        return self.t


class PadArrays(CtrlNode):
    """Pad 1-D numpy arrays in a particular column"""
    nodeName = 'PadArrays'
    uiTemplate = [('data_column', 'combo', {}),
                  # ('output_size', 'intSpin', {'min': -1, 'max': 9999999, 'step': 100, 'value': -1}),
                  ('method', 'combo', {'items': ['fill-size', 'random']}),
                  ('mode', 'combo', {'items': ['minimum', 'constant', 'edge', 'maximum',
                                               'mean', 'median', 'reflect', 'symmetric', 'wrap'], 'toolTip': 'Passed to numpy.pad "mode" parameter'}),
                  ('constant', 'doubleSpin', {'min': -9999999.9, 'max': 9999999.9, 'value': 1.0, 'step': 10.0,
                                              'tooltip': 'Value to use if "mode" is set to "constant"'}),
                  ('pad_with_nans', 'check', {'checked': False,
                                              'toolTip': 'Pad with nans, only if `mode` is set to `constant`'}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if not self.ctrls['Apply'].isChecked():
            return

        self.t = transmission.copy()

        a = self.t.df[self.data_column].values

        method = self.ctrls['method'].currentText()
        mode = self.ctrls['mode'].currentText()
        constant = self.ctrls['constant'].value()
        pad_with_nans = self.ctrls['pad_with_nans'].isChecked()

        if mode == 'constant' and pad_with_nans:
            constant = np.nan

        params = \
            {
                'method': method,
                'mode': mode,
                'constant': constant,
                'pad_with_nans': pad_with_nans
            }

        self.padded_output = self._pad(a, method, mode, constant, pad_with_nans)

        self.t.df['_PADDED'] = self.padded_output.tolist()
        self.t.df['_PADDED'] = self.t.df['_PADDED'].apply(np.array)

        self.t.history_trace.add_operation(data_block_id='all', operation='pad_arrays', parameters=params)
        self.t.last_output = '_PADDED'

        return self.t

    def _pad(self, a, method, mode, constant, pad_with_nans) -> np.ndarray:
        l = 0  # size of largest time series

        # Get size of largest time series
        for c in a:
            s = c.size
            if s > l:
                l = s

        # pre-allocate output array
        p = np.zeros(shape=(a.size, l))

        if mode == 'constant':
            kwargs = \
                {
                    'mode': 'constant',
                    'constant_values': constant
                }
        else:
            kwargs = {'mode': mode}

        # pad each 1D time series
        for i in tqdm(range(p.shape[0])):
            s = a[i].size

            if s == l:
                p[i, :] = a[i]
                continue

            max_pad_en_ix = l - s

            if method == 'random':
                pre = np.random.randint(0, max_pad_en_ix)
            elif method == 'fill-size':
                pre = 0
            else:
                raise ValueError('Must specific method as either "random" or "fill-size"')

            post = l - (pre + s)
            p[i, :] = np.pad(a[i], (pre, post), **kwargs)

        return p


class SelectRows(CtrlNode):
    pass


class SelectColumns(CtrlNode):
    pass


class TextFilter(CtrlNode):
    """Simple string filtering in a specified column"""
    nodeName = 'TextFilter'
    uiTemplate = [('Column', 'combo', {'toolTip': 'Filter according to this column'}),
                  ('filter', 'lineEdit', {'toolTip': 'Filter to apply in selected column'}),
                  ('Include', 'radioBtn', {'checked': True}),
                  ('Exclude', 'radioBtn', {'checked': False}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})]

    # def __init__(self, name):
    #     CtrlNode.__init__(self, name, terminals={'In': {'io': 'in'}, 'Out': {'io': 'out', 'bypass': 'In'}})
    #     self.ctrls['ROI_Type'].returnPressed.connect(self._setAvailTags)

    def processData(self, transmission: Transmission):
        ccols = organize_dataframe_columns(transmission.df.columns.to_list())[1]
        self.ctrls['Column'].setItems(ccols)
        col = self.ctrls['Column'].currentText()
        completer = QtWidgets.QCompleter(list(map(str, transmission.df[col].unique())))
        self.ctrls['filter'].setCompleter(completer)

        if self.ctrls['Apply'].isChecked() is False:
            return

        filt = self.ctrls['filter'].text()

        self.t = transmission.copy()

        include = self.ctrls['Include'].isChecked()
        exclude = self.ctrls['Exclude'].isChecked()

        params = {'column': col,
                  'filter': filt,
                  'include': include,
                  'exclude': exclude
                  }

        if include:
            self.t.df = self.t.df[self.t.df[col].astype(str) == filt]
        elif exclude:
            self.t.df = self.t.df[self.t.df[col].astype(str) != filt]

        self.t.df = self.t.df.reset_index(drop=True)

        self.t.history_trace.add_operation('all', operation='text_filter', parameters=params)

        return self.t


class NormRaw(CtrlNode):
    """Normalize between raw min and max values."""
    nodeName = 'NormRaw'
    uiTemplate = [('option', 'combo', {'items': ['top_5', 'top_10', 'top_5p', 'top_10p', 'top_25p', 'full_mean']}),
                  ('Apply', 'check', {'applyBox': True, 'checked': False})
                  ]

    def processData(self, transmission: Transmission):
        t = transmission
        dcols = organize_dataframe_columns(t.df.columns)[0]
        if not self.ctrls['Apply'].isChecked():
            return

        self.t = transmission.copy()

        output_column = '_NORMRAW'

        self.option = self.ctrls['option'].currentText()

        params = {'data_column': '_RAW_CURVE',
                  'option': self.option,
                  'output_column': output_column}

        self.proj_path = self.t.get_proj_path()

        tqdm().pandas()

        self.excluded = 0

        self.t.df[output_column] = self.t.df.progress_apply(lambda r: self._func(r['_RAW_CURVE'], r['ImgInfoPath'], r['ROI_State']), axis=1)

        self.t.history_trace.add_operation('all', 'normrawminmax', params)
        self.t.last_output = output_column

        if self.excluded > 0:
            QtWidgets.QMessageBox.warning(None, 'Curves excluded',
                                          f'The following number of curves were excluded because '
                                          f'the raw min value was larger than the max\n{self.excluded}')

            self.t.df = self.t.df[~self.t.df[output_column].isna()]

        return self.t

    def _func(self, data: np.ndarray, img_info_path: str, roi_state: dict) -> np.ndarray:
        if 'raw_min_max' in roi_state.keys():
            raw_min_max = roi_state['raw_min_max']

        else:
            cnmf_idx = roi_state['cnmf_idx']
            img_info_path = os.path.join(self.proj_path, img_info_path)
            roi_states = pickle.load(open(img_info_path, 'rb'))['roi_states']

            idx_components = roi_states['cnmf_output']['idx_components']

            list_ix = np.argwhere(idx_components == cnmf_idx).ravel().item()

            state = roi_states['states'][list_ix]

            if not state['cnmf_idx'] == cnmf_idx:
                raise ValueError('cnmf_idx from ImgInfoPath dict and DataFrame ROI_State dict do not match.')

            raw_min_max = state['raw_min_max']

        raw_min = raw_min_max['raw_min'][self.option]
        raw_max = raw_min_max['raw_max'][self.option]

        if raw_min >= raw_max:
            self.excluded += 1
            return np.NaN

        return TimeSeriesScalerMinMax(value_range=(raw_min, raw_max)).fit_transform(data).ravel()



from .data_types import Transmission
from .utils import *
from .math import cross_correlation, drfft_dtw, tvregdiff
from . import clustering_metrics
from . import meta_clustering
from .batch_manager_reader import load_batch, CaimanExtensions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Kushal Kolar

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from fcsugar import Container
from .. import Transmission
import os
import pandas as pd


class TransmissionContainer(Transmission, Container):
    def __init__(self, *args, **kwargs):
        Transmission.__init__(self, *args, **kwargs)
        Container.__init__(self)

    @property
    def dataframe(self):
        proj_df_path = os.path.join(self.get_proj_path(), 'dataframes', 'root.dfr')
        proj_df = pd.read_hdf(proj_df_path, key='project_dataframe', mode='r')
        return pd.merge(proj_df, self.df, on='uuid_curve')

    def append_log(self, node):
        super(TransmissionContainer, self).append_log(node)
        self.history_trace.add_operation('all', node.name, node.params)

    def __add__(self, other):
        if not isinstance(other, TransmissionContainer):
            raise TypeError("Can only combine TransmissionContainers")

        return TransmissionContainer.merge([self, other])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 17 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from .pytemplates.column_pytemplate import Ui_column_template
from functools import partial
import pandas as pd
from numpy import int64, float64
from ...common import configuration, get_project_manager


class ColumnWidget(QtWidgets.QWidget, Ui_column_template):
    signal_apply_clicked = QtCore.pyqtSignal(dict)
    signal_sample_id = QtCore.pyqtSignal(str)
    signal_items_selected = QtCore.pyqtSignal(list)
    signal_sample_id_delete_request = QtCore.pyqtSignal(str)

    def __init__(self, parent, tab_name: str, column_name: str, is_root=False):
        # super(ColumnWidget, self).__init__(parent)
        QtWidgets.QWidget.__init__(self, parent=parent)
        # Ui_column_template.__init__(self)
        self.setupUi(self)

        self.tab_name = tab_name
        self.column_name = column_name

        self.btnApply.clicked.connect(partial(self.btn_apply_clicked_emit_dict, option=None))
        self.btnReset.clicked.connect(self.lineEdit.clear)

        self.listWidget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.listWidget.itemSelectionChanged.connect(self.set_lineEdit)

        # self.listWidget.itemSelectionChanged.connect(self.emit_highlighted_items)

        if self.column_name == 'SampleID':
            self.listWidget.itemDoubleClicked.connect(self.emit_sample_id)
            if is_root:
                action_delete_sample = QtWidgets.QWidgetAction(self)
                action_delete_sample.setText('Delete sample')
                action_delete_sample.triggered.connect(self.delete_sample_id)

                self.delete_menu = QtWidgets.QMenu(self)
                self.delete_menu.addAction(action_delete_sample)

                self.listWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
                self.listWidget.customContextMenuRequested.connect(self.delete_context_menu_requested)
        else:
            self.listWidget.itemDoubleClicked.connect(partial(self.btn_apply_clicked_emit_dict, option='new'))

        if not is_root:
            self.btnApply.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.btnApply.customContextMenuRequested.connect(self.btnApply_context_menu_requested)

            apply_all = QtWidgets.QWidgetAction(self)
            apply_all.setText('Apply all')
            apply_all.triggered.connect(partial(self.btn_apply_clicked_emit_dict, option='all'))

            apply_all_in_new_tab = QtWidgets.QWidgetAction(self)
            apply_all_in_new_tab.setText('Apply all in new tab')
            apply_all_in_new_tab.triggered.connect(partial(self.btn_apply_clicked_emit_dict, option='all_in_new'))

            apply_in_new_tab = QtWidgets.QWidgetAction(self)
            apply_in_new_tab.setText('Apply in new tab')
            apply_in_new_tab.triggered.connect(partial(self.btn_apply_clicked_emit_dict, option='new'))

            self.btnApplyMenu = QtWidgets.QMenu(self)
            self.btnApplyMenu.addAction(apply_all)
            self.btnApplyMenu.addAction(apply_all_in_new_tab)
            self.btnApplyMenu.addSeparator()
            self.btnApplyMenu.addAction(apply_in_new_tab)

        self.lineEdit.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.lineEdit.customContextMenuRequested.connect(self.lineEdit_context_menu_requested)

        lineEdit_view_all = QtWidgets.QWidgetAction(self)
        lineEdit_view_all.setText('...')
        lineEdit_view_all.triggered.connect(self.line_edit_view_all_in_text_browser)

        lineEditNegation = QtWidgets.QWidgetAction(self)
        lineEditNegation.setText('"Not" modifer $NOT:')
        lineEditNegation.setToolTip('Perform negation of filter entries')
        lineEditNegation.triggered.connect(partial(self.set_modifier, '$NOT:'))

        lineEditSTR = QtWidgets.QWidgetAction(self)
        lineEditSTR.setText('"String" modifer $STR:')
        lineEditSTR.setToolTip('Treat filter entries as a string (as normal text)')
        lineEditSTR.triggered.connect(partial(self.set_modifier, '$STR:'))

        lineEditSTREQ = QtWidgets.QWidgetAction(self)
        lineEditSTREQ.setText('"String equals" modifer (exact match of text) $STR=:')
        lineEditSTREQ.setToolTip('Exact match of filter entries')
        lineEditSTREQ.triggered.connect(partial(self.set_modifier, '$STR=:'))

        lineEditSTRNOTEQ = QtWidgets.QWidgetAction(self)
        lineEditSTRNOTEQ.setText('"String not equals" modifer $STR!=:')
        lineEditSTRNOTEQ.setToolTip('Negation of exact match of filter entries')
        lineEditSTRNOTEQ.triggered.connect(partial(self.set_modifier, '$STR!=:'))

        self.lineEditMenu = QtWidgets.QMenu(self)
        self.lineEditMenu.addAction(lineEdit_view_all)
        self.lineEditMenu.addAction(lineEditNegation)
        self.lineEditMenu.addAction(lineEditSTR)
        self.lineEditMenu.addAction(lineEditSTREQ)
        self.lineEditMenu.addAction(lineEditSTRNOTEQ)

    def delete_context_menu_requested(self, p):
        self.delete_menu.exec_(self.listWidget.mapToGlobal(p))

    def delete_sample_id(self):
        sample_id = self.listWidget.currentItem().text()
        if QtWidgets.QMessageBox.question(self,
                                          'Delete sample',
                                          'Are you sure you want to delete sample id: "' + sample_id + '" ?\n',
                                          QtWidgets.QMessageBox.Yes,
                                          QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.No:
            return
        get_project_manager().delete_sample_id_rows(sample_id)

    @property
    def series(self) -> pd.Series:
        return self._series

    @series.setter
    def series(self, series: pd.Series):
        self._series = series
        if len(series) < 1:
            self.set_empty()
            return

        self.listWidget.clear()
        self.setEnabled(True)

        self.column_type = type(self.series[self.series.index.values.min()])

        if self.column_type is str:
            self.set_as_str()
        elif self.column_type in [int, float, int64, float64]:
            self.set_as_num()
        elif self.column_type is list:
            self.set_as_list()
        elif self.column_type is bool:
            self.set_as_bool()
        else:
            self.set_as_unknown()

        if self.column_name == 'SampleID':
            self.labelColumnName.setText('SampleID: n = ' + str(self.listWidget.count()) + '\n' + str(self.column_type))
        else:
            self.labelColumnName.setText(self.column_name + '\n' + str(self.column_type))

    def set_empty(self):
        self.listWidget.clear()
        self.listWidget.addItems(['Empty'])
        self.setDisabled(True)

    def btn_apply_clicked_emit_dict(self, option: str = None):
        d = {'column_widget': self, 'option': option}

        self.signal_apply_clicked.emit(d)
        print(d)

    def btnApply_context_menu_requested(self, p):
        self.btnApplyMenu.exec_(self.btnApply.mapToGlobal(p))

    def lineEdit_context_menu_requested(self, p):
        self.lineEditMenu.exec_(self.lineEdit.mapToGlobal(p))

    def line_edit_view_all_in_text_browser(self):
        pass

    def set_modifier(self, modifier: str):
        text = self.lineEdit.text()
        self.lineEdit.setText(modifier + text)

    def set_as_str(self):
        l = list(set(self.series))
        l.sort()
        self.listWidget.addItems(l)

        lineEdit_exact_match = QtWidgets.QWidgetAction(self)
        lineEdit_exact_match.setText('Exact match')

    def set_as_num(self):
        l = [str(x) for x in set(self.series)]
        l.sort()
        self.listWidget.addItems(l)

        lineEdit_greater_than = QtWidgets.QWidgetAction(self)
        lineEdit_greater_than.setText('Greater than')
        lineEdit_greater_than.triggered.connect(partial(self.set_num_modifier, '$>:'))
        self.lineEditMenu.addAction(lineEdit_greater_than)

        lineEdit_less_than = QtWidgets.QWidgetAction(self)
        lineEdit_less_than.setText('Less than')
        lineEdit_less_than.triggered.connect(partial(self.set_num_modifier, '$<:'))
        self.lineEditMenu.addAction(lineEdit_less_than)

        lineEdit_less_than_eq = QtWidgets.QWidgetAction(self)
        lineEdit_less_than_eq.setText('Less than or equal to')
        lineEdit_less_than_eq.triggered.connect(partial(self.set_num_modifier, '$<=:'))
        self.lineEditMenu.addAction(lineEdit_less_than_eq)

        lineEdit_greater_than_eq = QtWidgets.QWidgetAction(self)
        lineEdit_greater_than_eq.setText('Greater than or equal to')
        lineEdit_greater_than_eq.triggered.connect(partial(self.set_num_modifier, '$>=:'))
        self.lineEditMenu.addAction(lineEdit_greater_than_eq)

    def set_num_modifier(self, modifier: str):
        text = self.lineEdit.text()
        self.lineEdit.setText(modifier + text)

    def set_as_bool(self):
        l = [str(x) for x in set(self.series)]
        l.sort()
        self.listWidget.addItems(l)

    def set_as_list(self):
        l = list(set([a for b in self.series.tolist() for a in b]))
        l.sort()
        self.listWidget.addItems(l)

        lineEdit_exact_match = QtWidgets.QWidgetAction(self)
        lineEdit_exact_match.setText('Exact match')

    def set_as_unknown(self):
        self.listWidget.addItems(['Unsupported type'])
        self.setDisabled(True)

    def set_lineEdit(self):
        txt = ''
        for item in self.listWidget.selectedItems():
            txt = txt + '|' + item.text()
        txt = txt[1:]
        self.lineEdit.setText(txt)

    def emit_highlighted_items(self):
        self.signal_items_selected.emit(self.listWidget.selectedItems())

    def emit_sample_id(self, item: QtWidgets.QListWidgetItem):
        self.signal_sample_id.emit(item.text())

        # def clear(self):
        #     self.listWidget.clear()
        #     self.listWidget.setEnabled(True)


if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    df = pd.read_pickle('/home/kushal/Sars_stuff/github-repos/testprojects/March 9/dataframes/root.mzp')

    c = ColumnWidget(parent=None, tab_name='bah_tab', column_name='Genotype', series=df['odor'])

    c.show()
    app.exec_()
